# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for unified input/output
"""

from astropy.utils.compat.gzip import GzipFile

from glue.lal import CacheEntry

from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def identify_factory(*extensions):
    def identify(*args, **kwargs):
        """Identify the given extensions in a file object/path
        """
        fp = args[3]
        if isinstance(fp, (file, GzipFile)):
            fp = fp.name
        elif isinstance(fp, CacheEntry):
            fp = fp.path
        # identify string
        if isinstance(fp, (unicode, str)) and fp.endswith(extensions):
            return True
        else:
            return False
    return identify


def gopen(name, *args, **kwargs):
    """Open a file handling optional gzipping

    Parameters
    ----------
    name : `str`
        path (name) of file to open
    *args, **kwargs
        other arguments to pass to either `open` for regular files, or
        `gzip.open` for files with a `name` ending in `.gz`
    """
    if name.endswith('.gz'):
        return GzipFile(name, *args, **kwargs)
    else:
        return open(name, *args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output routines for gravitational-wave frame (GWF) format files.

The frame format is defined in LIGO-T970130 available from dcc.ligo.org.

Currently supported are two separate libraries:

- `lalframe` : using the LIGO Algorithm Library Frame API (based off the
  FrameL library)
- `framecpp` : using the alternative ``frameCPP`` library

Due to the lower-level nature of the frameCPP python package, it is
preferred, in the instance that both lalframe and frameCPP are available
on a system.
"""

import importlib

from astropy.io.registry import register_reader

from ....utils import with_import
from ....version import version
from ... import (TimeSeries, TimeSeriesDict, StateVector, StateVectorDict)

from .identify import *

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version

BUILTIN_LIBRARIES = [
    'lalframe',
    'framecpp',
]


def channel_dict_kwarg(value, channels, types=None, astype=None):
    """Format the given kwarg value in a dict with one value per channel

    Parameters
    ----------
    value : any type
        keyword argument value as given by user
    channels : `list`
        list of channels being read
    types : `list` of `type`
        list of valid object types for value
    astype : `type`
        output type for `dict` values

    Returns
    -------
    dict : `dict`
        `dict` of values, one value per channel key, if parsing is successful
    None : `None`
        `None`, if parsing was unsuccessful
    """
    if types is not None and isinstance(value, tuple(types)):
        out = dict((c, value) for c in channels)
    elif isinstance(value, (tuple, list)):
        out = dict(zip(channels, value))
    elif value is None:
        out = dict()
    elif isinstance(value, dict):
        out = value.copy()
    else:
        return None
    if astype is not None:
        return dict((key, astype(out[key])) for key in out)
    else:
        return out


def register_gwf_io_library(library, package='gwpy.timeseries.io.gwf'):
    """Register a full set of GWF I/O methods for the given library

    The given library must define a `read_timeseriesdict`
    method, taking a GWF source and a list of channels, and kwargs,
    and returning a `TimeSeriesDict` of data.

    Additionally, the library must store the name of the upstream I/O
    library it uses in the 'DEPENDS' variable.

    This method then wraps the `read_timeseriesdict` method from the
    given to provide readers for the `TimeSeries`, `StateVector`, and
    `StateVectorDict` objects
    """
    # import library
    lib = importlib.import_module('.%s' % library, package=package)
    dependency = lib.DEPENDS
    reader = lib.read_timeseriesdict

    def read_timeseriesdict(source, *args, **kwargs):
        """Read `TimeSeriesDict` from GWF source
        """
        if isinstance(source, list) and not len(source):
            raise ValueError("Cannot read from empty %s"
                             % type(source).__name__)
        # use multiprocessing or padding
        nproc = kwargs.pop('nproc', 1)
        pad = kwargs.pop('pad', None)
        if nproc > 1 or pad is not None:
            from ..cache import read_cache
            kwargs['target'] = TimeSeriesDict
            kwargs['nproc'] = nproc
            kwargs['pad'] = pad
            return read_cache(source, *args, **kwargs)
        else:
            return reader(source, *args, **kwargs)

    @with_import(dependency)
    def read_timeseries(source, channel, *args, **kwargs):
        """Read `TimeSeries` from GWF source
        """
        return read_timeseriesdict(source, [channel], *args, **kwargs)[channel]

    @with_import(dependency)
    def read_statevector(source, channel, *args, **kwargs):
        """Read `StateVector` from GWF source
        """
        bits = kwargs.pop('bits', None)
        kwargs.setdefault('_SeriesClass', StateVector)
        sv = read_timeseries(source, channel, *args, **kwargs)
        sv.bits = bits
        return sv

    @with_import(dependency)
    def read_statevectordict(source, channels, *args, **kwargs):
        """Read `StateVectorDict` from GWF source
        """
        bitss = kwargs.pop('bits', {})
        # read list of bit lists
        if (isinstance(bitss, (list, tuple)) and len(bitss) and
                isinstance(bitss[0], (list, tuple))):
            bitss = dict(zip(channels, bitss))
        # read single list for all channels
        elif isinstance(bitss, (list, tuple)):
            bitss = dict((channel, bitss) for channel in channels)
        # otherwise assume dict of bit lists

        # read data as timeseriesdict and repackage with bits
        kwargs.setdefault('_SeriesClass', StateVector)
        svd = StateVectorDict(
            read_timeseriesdict(source, channels, *args, **kwargs))
        for (channel, bits) in bitss.iteritems():
            svd[channel].bits = bits
        return svd

    # register format
    register_reader(library, TimeSeriesDict, read_timeseriesdict)
    register_reader(library, TimeSeries, read_timeseries)
    register_reader(library, StateVectorDict, read_statevectordict)
    register_reader(library, StateVector, read_statevector)

    # register generic 'GWF' format
    try:
        __import__(dependency, fromlist=[''])
    except ImportError:
        pass
    else:
        register_reader('gwf', TimeSeriesDict, read_timeseriesdict,
                        force=True)
        register_reader('gwf', TimeSeries, read_timeseries,
                        force=True)
        register_reader('gwf', StateVectorDict, read_statevectordict,
                        force=True)
        register_reader('gwf', StateVector, read_statevector,
                        force=True)
    return (read_timeseries, read_timeseriesdict,
            read_statevector, read_statevectordict)


# register builtin libraries
for library in BUILTIN_LIBRARIES:
    register_gwf_io_library(library)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""A collection of average power spectral density calculation routines

This module defines a set of builtin methods with which to calculate
an average power spectral density. Users can define their own through
the registry method `register_method`.
"""

from ..utils.compat import OrderedDict
from .. import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

spectrum_methods = OrderedDict()
density_methods = OrderedDict()


def register_method(func, name=None, force=False, scaling='density'):
    """Register a method of calculating an average spectrogram.

    Parameters
    ----------
    func : `callable`
        function to execute
    name : `str`, optional
        name of the method, defaults to ``func.__name__``
    """
    from ..timeseries import TimeSeries
    # get name and format doc addition
    if name is None:
        name = func.__name__
    path = '.'.join([func.__module__, func.__name__])
    doc = '        - %r: :meth:`%s`' % (name, path)

    def append_doc(obj, newdoc):
        lines = obj.__doc__.splitlines()
        if not lines[-1].strip(' '):
            suffix = lines.pop(-1)
        else:
            suffix = ''
        obj.__func__.__doc__ = '\n'.join(lines + [newdoc, suffix])

    # determine spectrum method type, and append doc
    if scaling == 'density':
        methods = density_methods
        append_doc(TimeSeries.psd, doc)
        append_doc(TimeSeries.asd, doc)
    elif scaling == 'spectrum':
        methods = spectrum_methods
        append_doc(TimeSeries.power_spectrum, doc)
    else:
        raise ValueError("Unknown Spectrum type: %r" % scaling)
    # record method in registry
    if name in methods and not force:
        raise KeyError("'%s' already registered, use force=True to override."
                       % name)
    methods[name] = func


def get_method(name, scaling='density'):
    """Return the Specrum generator registered with the given name.
    """
    if scaling == 'density':
        try:
            return density_methods[name]
        except KeyError:
            raise ValueError("No PSD method registered with name %r" % name)
    elif scaling == 'spectrum':
        try:
            return spectrum_methods[name]
        except KeyError:
            raise ValueError("No power spectrum method registered with "
                             "name %r" % name)
    else:
        raise ValueError("Unknown Spectrum type: %r" % scaling)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read LIGO_LW documents into glue.ligolw.table.Table objects.
"""

from astropy.io import registry

from glue.ligolw.table import StripTableName as strip
from glue.ligolw.lsctables import TableByName

from ...io.ligolw import (table_from_file, identify_ligolw)
from ... import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version
__all__ = []


def read_table_factory(table_):
    """Define a custom function to read this table from a LIGO_LW file.
    """
    def _read(f, *args, **kwargs):
        return table_from_file(f, table_.tableName, *args, **kwargs)
    return _read

# register reader and auto-id for LIGO_LW
for table in TableByName.itervalues():
    tablename = strip(table.tableName)
    func = read_table_factory(table)
    # register generic reader and table-specific reader
    registry.register_reader('ligolw', table, func)
    registry.register_reader(tablename, table, func)
    registry.register_identifier('ligolw', table, identify_ligolw)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for the GWpy command-line interface (CLI)
"""

import argparse

from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class GPSAction(argparse.Action):
    """Custom `~argparse.Action` to convert date strings (or times) to GPS

    The input values are passed to the `~gwpy.time.to_gps` method which
    will return a `~gwpy.time.LIGOTimeGPS` output.

    See Also
    --------
    argparse.Action
        for details on how to use this object
    """
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            values = float(values)
        except (TypeError, ValueError):
            pass
        setattr(namespace, self.dest, to_gps(values))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013-2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module registers a number of custom units used in GW astronomy.
"""

from .. import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

from astropy import units

# enable imperial units
units.add_enabled_units(units.imperial)

# -----------------------------------------------------------------------------
# instrumental units

units.add_enabled_units([
    units.def_unit(['counts'], represents=units.Unit('count')),
    units.def_unit(['undef'], doc='No unit has been defined for these data'),
    units.def_unit(['coherence'], represents=units.dimensionless_unscaled),
    units.def_unit(['strain'], represents=units.dimensionless_unscaled),
    units.def_unit(['Degrees_C'], represents=units.Unit('Celsius')),
    units.def_unit(['Degrees_F'], represents=units.Unit('Fahrenheit')),
])



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for plotter module
"""

import os
import tempfile
from math import pi

import numpy
from numpy import testing as nptest

from scipy import signal

from matplotlib import use
use('agg')
from matplotlib.legend import Legend
from matplotlib.colors import LogNorm
from matplotlib.collections import PatchCollection

from astropy import units

from compat import unittest

from gwpy import version
from gwpy.segments import (DataQualityFlag, Segment, SegmentList)
from gwpy.timeseries import TimeSeries
from gwpy.plotter import (figure, Plot, Axes,
                          TimeSeriesPlot, TimeSeriesAxes,
                          SpectrumPlot, SpectrumAxes,
                          EventTablePlot, EventTableAxes,
                          HistogramPlot, HistogramAxes,
                          SegmentPlot, SegmentAxes,
                          SpectrogramPlot, BodePlot)
from gwpy.plotter.gps import (GPSTransform, InvertedGPSTransform)
from gwpy.plotter.html import map_data
from gwpy.plotter.tex import (float_to_latex, label_to_latex,
                              unit_to_latex, USE_TEX)

from test_timeseries import TEST_HDF_FILE

# design ZPK for BodePlot test
ZPK = [100], [1], 1e-2
FREQUENCIES, H = signal.freqresp(ZPK, n=200)
FREQUENCIES /= 2. * pi
MAGNITUDE = 20 * numpy.log10(numpy.absolute(H))
PHASE = numpy.degrees(numpy.unwrap(numpy.angle(H)))

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class Mixin(object):
    FIGURE_CLASS = Plot
    AXES_CLASS = Axes

    def new(self):
        """Create a new `Figure` with some `Axes`

        Returns (fig, ax)
        """
        fig = self.FIGURE_CLASS()
        return fig, fig.gca()


class PlotTestCase(Mixin, unittest.TestCase):
    """`TestCase` for the `gwpy.plotter` module
    """
    def test_init(self):
        # test object creation
        fig, ax = self.new()
        self.assertIsInstance(fig, self.FIGURE_CLASS)
        # test added properties
        self.assertIsInstance(fig.colorbars, list)
        self.assertEqual(len(fig.colorbars), 0)
        self.assertIsInstance(fig._coloraxes, list)
        self.assertEqual(len(fig._coloraxes), 0)
        # test gca
        self.assertIsInstance(ax, self.AXES_CLASS)
        fig.close()

    def test_auto_refresh(self):
        fig = self.FIGURE_CLASS()
        self.assertFalse(fig.get_auto_refresh())
        fig.set_auto_refresh(True)
        self.assertTrue(fig.get_auto_refresh())
        fig = self.FIGURE_CLASS(auto_refresh=True)
        self.assertTrue(fig.get_auto_refresh())
        fig.close()

    def test_save(self):
        fig = self.FIGURE_CLASS()
        f = tempfile.mktemp(suffix='.png')
        try:
            fig.save(f)
        finally:
            if os.path.isfile(f):
                os.remove(f)
            fig.close()

    # -- test axes_method decorators
    def test_axes_methods(self):
        fig, ax = self.new()
        # get_xlim
        self.assertTupleEqual(fig.get_xlim(), ax.get_xlim())
        # set_xlim
        fig.set_xlim(4, 5)
        self.assertTupleEqual(fig.get_xlim(), (4, 5))
        self.assertTupleEqual(fig.get_xlim(), ax.get_xlim())

        # get_xlabel
        self.assertEqual(fig.get_xlabel(), ax.get_xlabel())
        # set_xlabel
        fig.set_xlabel('test')
        self.assertEqual(fig.get_xlabel(), 'test')
        self.assertEqual(fig.get_xlabel(), ax.get_xlabel())

        # get_ylim
        self.assertTupleEqual(fig.get_ylim(), ax.get_ylim())
        # set_ylim
        fig.set_ylim(4, 5)
        self.assertTupleEqual(fig.get_ylim(), (4, 5))
        self.assertTupleEqual(fig.get_ylim(), ax.get_ylim())

        # get_ylabel
        self.assertEqual(fig.get_ylabel(), ax.get_ylabel())
        # set_ylabel
        fig.set_ylabel('test')
        self.assertEqual(fig.get_ylabel(), 'test')
        self.assertEqual(fig.get_ylabel(), ax.get_ylabel())

        # title
        self.assertEqual(fig.get_title(), ax.get_title())
        # set_title
        fig.set_title('Test')
        self.assertEqual(fig.get_title(), 'Test')
        self.assertEqual(fig.get_title(), ax.get_title())

        # xscale
        self.assertEqual(fig.get_xscale(), ax.get_xscale())
        # set_xscale
        fig.set_xscale('log')
        self.assertEqual(fig.get_xscale(), 'log')
        self.assertEqual(fig.get_xscale(), ax.get_xscale())

        # yscale
        self.assertEqual(fig.get_yscale(), ax.get_yscale())
        # set_yscale
        fig.set_yscale('log')
        self.assertEqual(fig.get_yscale(), 'log')
        self.assertEqual(fig.get_yscale(), ax.get_yscale())
        fig.close()

    def test_logx(self):
        fig, ax = self.new()
        fig.set_xlim(0.1, 10)
        fig.logx = True
        self.assertEqual(ax.get_xscale(), 'log')
        self.assertTrue(fig.logx)
        fig.close()

    def test_logy(self):
        fig, ax = self.new()
        fig.set_ylim(0.1, 10)
        fig.logy = True
        self.assertEqual(ax.get_yscale(), 'log')
        self.assertTrue(fig.logy)
        fig.close()

    def test_add_legend(self):
        fig, ax = self.new()
        self.assertIsNone(fig.add_legend())
        ax.plot([1, 2, 3, 4], label='Plot')
        self.assertIsInstance(fig.add_legend(), Legend)
        fig.close()

    def test_figure(self):
        fig = figure()
        self.assertIsInstance(fig, Plot)
        fig = figure(FigureClass=self.FIGURE_CLASS)
        self.assertIsInstance(fig, self.FIGURE_CLASS)


class AxesTestCase(Mixin, unittest.TestCase):
    def test_legend(self):
        fig = self.FIGURE_CLASS()
        ax = fig.add_subplot(111, projection=self.AXES_CLASS.name)
        ax.plot([1, 2, 3, 4], label='Plot')
        leg = ax.legend()
        self.assertEqual(leg.get_frame().get_alpha(), 0.8)
        for l in leg.get_lines():
            self.assertEqual(l.get_linewidth(), 8)
        fig.close()


# -- TimeSeries plotters ------------------------------------------------------

class TimeSeriesMixin(object):
    FIGURE_CLASS = TimeSeriesPlot
    AXES_CLASS = TimeSeriesAxes

    def setUp(self):
        self.ts = TimeSeries.read(TEST_HDF_FILE, 'H1:LDAS-STRAIN')
        self.sg = self.ts.spectrogram2(0.5, 0.49)
        self.mmm = [self.ts, self.ts * 0.9, self.ts*1.1]


class TimeSeriesPlotTestCase(TimeSeriesMixin, PlotTestCase):
    def test_init(self):
        # test empty
        fig, ax = self.new()
        self.assertIsInstance(ax, self.AXES_CLASS)
        # test passing arguments
        fig = self.FIGURE_CLASS(figsize=[9, 6])
        self.assertEqual(fig.get_figwidth(), 9)
        fig = self.FIGURE_CLASS(self.ts)
        ax = fig.gca()
        self.assertEqual(len(ax.lines), 1)
        self.assertEqual(fig.get_epoch(), self.ts.x0.value)
        self.assertEqual(fig.get_xlim(), self.ts.span)
        # test passing multiple timeseries
        fig = self.FIGURE_CLASS(self.ts, self.ts)
        self.assertEqual(len(fig.gca().lines), 2)
        fig = self.FIGURE_CLASS(self.ts, self.ts, sep=True, sharex=True)
        self.assertEqual(len(fig.axes), 2)
        for ax in fig.axes:
            self.assertEqual(len(ax.lines), 1)
        self.assertIs(fig.axes[1]._sharex, fig.axes[0])

    def test_add_colorbar(self):
        def make_fig():
            fig = self.FIGURE_CLASS(self.sg ** (1/2.))
            return fig, fig.gca()

        # test basic
        fig, ax = make_fig()
        cb = fig.add_colorbar()
        self.assertEqual(len(fig.colorbars), 1)
        self.assertIs(cb, fig.colorbars[0])
        fig.close()
        # test kwargs
        fig, ax = make_fig()
        fig.add_colorbar(ax=ax, norm='log', clim=[1e-22, 1e-18],
                         label='Test colorbar')
        fig.close()


class TimeSeriesAxesTestCase(TimeSeriesMixin, AxesTestCase):
    def test_init(self):
        fig, ax = self.new()
        self.assertIsInstance(ax, self.AXES_CLASS)
        self.assertEqual(ax.get_epoch(), 0)
        self.assertEqual(ax.get_xscale(), 'auto-gps')
        self.assertEqual(ax.get_xlabel(), '_auto')
        fig.close()

    def test_plot_timeseries(self):
        fig, ax = self.new()
        # check method works
        ax.plot_timeseries(self.ts)
        # check data are correct
        l = ax.get_lines()[0]
        nptest.assert_array_equal(l.get_xdata(), self.ts.times.value)
        nptest.assert_array_equal(l.get_ydata(), self.ts.value)
        # check GPS axis is set ok
        self.assertEqual(ax.get_epoch(), self.ts.x0.value)
        self.assertTupleEqual(ax.get_xlim(), tuple(self.ts.span))
        fig.close()

    def test_plot_timeseries_mmm(self):
        fig, ax = self.new()
        # test default
        artists = ax.plot_timeseries_mmm(*self.mmm)
        self.assertEqual(len(artists), 5)
        self.assertEqual(len(ax.lines), 3)
        self.assertEqual(len(ax.collections), 2)
        fig.close()
        # test min only
        fig, ax = self.new()
        artists = ax.plot_timeseries_mmm(self.mmm[0], min_=self.mmm[1])
        self.assertEqual(len(artists), 5)
        self.assertIsNone(artists[3])
        self.assertIsNone(artists[4])
        self.assertEqual(len(ax.lines), 2)
        self.assertEqual(len(ax.collections), 1)
        fig.close()
        # test max only
        fig, ax = self.new()
        artists = ax.plot_timeseries_mmm(self.mmm[0], max_=self.mmm[2])
        self.assertEqual(len(artists), 5)
        self.assertIsNone(artists[1])
        self.assertIsNone(artists[2])
        self.assertEqual(len(ax.lines), 2)
        self.assertEqual(len(ax.collections), 1)

    def test_plot_spectrogram(self):
        fig, ax = self.new()
        # check method
        ax.plot_spectrogram(self.sg)
        coll = ax.collections[0]
        nptest.assert_array_equal(coll.get_array(), self.sg.value.T.flatten())
        # check GPS axis is set ok
        self.assertEqual(ax.get_epoch(), self.sg.x0.value)
        self.assertTupleEqual(ax.get_xlim(), tuple(self.sg.xspan))
        # check frequency axis
        if USE_TEX:
            self.assertEqual(ax.get_ylabel(), r'Frequency [$\mathrm{Hz}$]')
        else:
            self.assertEqual(ax.get_ylabel(), r'Frequency [Hz]')
        # check kwarg parsing
        c = ax.plot_spectrogram(self.sg, norm='log')
        self.assertIsInstance(c.norm, LogNorm)
        fig.close()

    def test_plot(self):
        fig, ax = self.new()
        ax.plot(self.ts)
        fig.close()


# -- Spectrum plotters --------------------------------------------------------

class SpectrumMixin(object):
    FIGURE_CLASS = SpectrumPlot
    AXES_CLASS = SpectrumAxes

    def setUp(self):
        self.ts = TimeSeries.read(TEST_HDF_FILE, 'H1:LDAS-STRAIN')
        self.asd = self.ts.asd(1)
        self.mmm = [self.asd, self.asd*0.9, self.asd*1.1]

class SpectrumPlotTestCase(SpectrumMixin, PlotTestCase):
    def test_init(self):
        super(SpectrumPlotTestCase, self).test_init()
        # test convenience plotting
        fig = self.FIGURE_CLASS(self.asd)
        self.assertEqual(len(fig.axes), 1)
        ax = fig.gca()
        self.assertIsInstance(ax, self.AXES_CLASS)
        self.assertEqual(len(ax.lines), 1)


class SpectrumAxesTestCase(SpectrumMixin, AxesTestCase):
    def test_plot_spectrum(self):
        fig, ax = self.new()
        l = ax.plot_spectrum(self.asd)[0]
        nptest.assert_array_equal(l.get_xdata(), self.asd.frequencies.value)
        nptest.assert_array_equal(l.get_ydata(), self.asd.value)
        fig.close()

    def test_plot_spectrum_mmm(self):
        fig, ax = self.new()
        # test defaults
        artists = ax.plot_spectrum_mmm(*self.mmm)
        self.assertEqual(len(artists), 5)
        self.assertEqual(len(ax.lines), 3)
        self.assertEqual(len(ax.collections), 2)
        fig.close()
        # test min only
        fig, ax = self.new()
        artists = ax.plot_spectrum_mmm(self.mmm[0], min_=self.mmm[1])
        self.assertEqual(len(artists), 5)
        self.assertIsNone(artists[3])
        self.assertIsNone(artists[4])
        self.assertEqual(len(ax.lines), 2)
        self.assertEqual(len(ax.collections), 1)
        fig.close()
        # test max only
        fig, ax = self.new()
        artists = ax.plot_spectrum_mmm(self.mmm[0], max_=self.mmm[2])
        self.assertEqual(len(artists), 5)
        self.assertIsNone(artists[1])
        self.assertIsNone(artists[2])
        self.assertEqual(len(ax.lines), 2)
        self.assertEqual(len(ax.collections), 1)


# -- Table plotters -----------------------------------------------------------

class EventTableMixin(object):
    FIGURE_CLASS = EventTablePlot
    AXES_CLASS = EventTableAxes


class EventTablePlotTestCase(EventTableMixin, PlotTestCase):
    pass


class EventTableAxesTestCase(EventTableMixin, AxesTestCase):
    pass


# -- Segment plotter ----------------------------------------------------------

class SegmentMixin(object):
    FIGURE_CLASS = SegmentPlot
    AXES_CLASS = SegmentAxes

    def setUp(self):
        self.segments = SegmentList([Segment(0, 3), Segment(6, 7)])
        active = SegmentList([Segment(1, 2), Segment(3, 4), Segment(5, 7)])
        self.flag = DataQualityFlag(name='Test segments', known=self.segments,
                                    active=active)


class SegmentPlotTestCase(SegmentMixin, PlotTestCase):
    pass


class SegmentAxesTestCase(SegmentMixin, AxesTestCase):

    def test_plot_dqflag(self):
        fig, ax = self.new()
        c = ax.plot_dqflag(self.flag)
        self.assertEqual(c.get_label(), self.flag.texname)
        self.assertEqual(len(ax.collections), 2)
        self.assertIs(ax.collections[1], c)

    def test_build_segment(self):
        patch = self.AXES_CLASS.build_segment((1.1, 2.4), 10)
        self.assertTupleEqual(patch.get_xy(), (1.1, 9.6))
        self.assertAlmostEqual(patch.get_height(), 0.8)
        self.assertAlmostEqual(patch.get_width(), 1.3)
        self.assertTupleEqual(patch.get_facecolor(), (0.0, 0.0, 1.0, 1.0))
        # check kwarg passing
        patch = self.AXES_CLASS.build_segment((1.1, 2.4), 10, facecolor='red')
        self.assertTupleEqual(patch.get_facecolor(), (1.0, 0.0, 0.0, 1.0))

    def test_plot_segmentlist(self):
        fig, ax = self.new()
        c = ax.plot_segmentlist(self.segments)
        self.assertIsInstance(c, PatchCollection)
        self.assertEqual(ax.dataLim.x0, 0.)
        self.assertAlmostEqual(ax.dataLim.x1, 7.)
        self.assertTupleEqual(ax.get_ylim(), (-1., 1.))
        self.assertEqual(len(c.get_paths()), len(self.segments))
        self.assertEqual(ax.get_epoch(), self.segments[0][0])
        # test y
        p = ax.plot_segmentlist(self.segments).get_paths()[0].get_extents()
        self.assertEqual(p.y0 + p.height/2., 1.)
        p = ax.plot_segmentlist(
            self.segments, y=8).get_paths()[0].get_extents()
        self.assertEqual(p.y0 + p.height/2., 8.)
        # test kwargs
        c = ax.plot_segmentlist(self.segments, label='My segments',
                                rasterized=True)
        self.assertEqual(c.get_label(), 'My segments')
        self.assertTrue(c.get_rasterized())
        fig.close()

    def test_plot(self):
        fig, ax = self.new()
        ax.plot(self.segments)
        ax.plot(self.flag)
        ax.plot(self.flag, self.segments)


# -- Histogram plotter --------------------------------------------------------

class HistogramMixin(object):
    FIGURE_CLASS = HistogramPlot
    AXES_CLASS = HistogramAxes

    def setUp(self):
        self.ts = TimeSeries.read(TEST_HDF_FILE, 'H1:LDAS-STRAIN')


class HistogramPlotTestCase(HistogramMixin, PlotTestCase):
    pass


class HistogramAxesTestCase(HistogramMixin, AxesTestCase):
    def test_hist_series(self):
        fig, ax = self.new()
        hist = ax.hist_series(self.ts)

    def test_hist(self):
        fig, ax = self.new()
        ax.hist(self.ts)

    def test_common_limits(self):
        fig, ax = self.new()
        a = ax.common_limits(self.ts.value * 1e16)
        self.assertTupleEqual(a, (-1.0227435293, 1.0975343221))
        b = ax.common_limits([self.ts.value * 1e16])
        self.assertEqual(a, b)

    def test_bin_boundaries(self):
        ax = self.AXES_CLASS
        nptest.assert_array_equal(ax.bin_boundaries(0, 10, 4),
                                  [0., 2.5, 5., 7.5, 10.])
        nptest.assert_array_equal(ax.bin_boundaries(1, 100, 2, log=True),
                                  [1, 10., 100.])
        self.assertRaises(ValueError, ax.bin_boundaries, 0, 100, 2, log=True)


# -- Filter plotter -----------------------------------------------------------

class BodePlotTestCase(unittest.TestCase):
    FIGURE_CLASS = BodePlot

    def test_init(self):
        fig = self.FIGURE_CLASS()
        self.assertEqual(len(fig.axes), 2)
        maxes, paxes = fig.axes
        # test magnigtude axes
        self.assertIsInstance(maxes, SpectrumAxes)
        self.assertEqual(maxes.get_xscale(), 'log')
        self.assertEqual(maxes.get_xlabel(), '')
        self.assertEqual(maxes.get_yscale(), 'linear')
        self.assertEqual(maxes.get_ylabel(), 'Magnitude [dB]')
        # test phase axes
        self.assertIsInstance(paxes, SpectrumAxes)
        self.assertEqual(paxes.get_xscale(), 'log')
        self.assertEqual(paxes.get_xlabel(), 'Frequency [Hz]')
        self.assertEqual(paxes.get_yscale(), 'linear')
        self.assertEqual(paxes.get_ylabel(), 'Phase [deg]')
        self.assertTupleEqual(paxes.get_ylim(), (-180, 180))

    def test_add_filter(self):
        # test method 1
        fig = self.FIGURE_CLASS()
        fig.add_filter(ZPK)
        lm = fig.maxes.get_lines()[0]
        lp = fig.paxes.get_lines()[0]
        nptest.assert_array_equal(lm.get_xdata(), FREQUENCIES)
        nptest.assert_array_equal(lm.get_ydata(), MAGNITUDE)
        nptest.assert_array_equal(lp.get_xdata(), FREQUENCIES)
        nptest.assert_array_equal(lp.get_ydata(), PHASE)
        fig.close()
        # test method 2
        fig = self.FIGURE_CLASS(ZPK)
        lm = fig.maxes.get_lines()[0]
        lp = fig.paxes.get_lines()[0]
        nptest.assert_array_equal(lm.get_xdata(), FREQUENCIES)
        nptest.assert_array_equal(lm.get_ydata(), MAGNITUDE)
        nptest.assert_array_equal(lp.get_xdata(), FREQUENCIES)
        nptest.assert_array_equal(lp.get_ydata(), PHASE)


# -- gwpy.plotter.gps module tests --------------------------------------------

class GpsTransformTestCase(unittest.TestCase):
    TRANSFORM = GPSTransform
    EPOCH = 100.0
    UNIT = 'minutes'
    SCALE = 60.
    X = 190.0
    A = 90.0
    B = 19/6.
    C = 1.5

    def test_empty(self):
        t = self.TRANSFORM()
        self.assertEqual(t.transform(1.0), 1.0)

    def test_epoch(self):
        transform = self.TRANSFORM(epoch=self.EPOCH)
        self.assertEqual(transform.get_epoch(), self.EPOCH)
        self.assertEqual(transform.transform(self.X), self.A)
        self.assertAlmostEqual(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)

    def test_scale(self):
        transform = self.TRANSFORM(unit=self.UNIT)
        self.assertEqual(transform.get_scale(), self.SCALE)
        self.assertEqual(transform.transform(self.X), self.B)
        self.assertAlmostEqual(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)

    def test_epoch_and_scale(self):
        transform = self.TRANSFORM(epoch=self.EPOCH, unit=self.UNIT)
        self.assertEqual(transform.transform(self.X), self.C)
        self.assertAlmostEqual(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)


class InverseGpsTransformTestCase(GpsTransformTestCase):
    TRANSFORM = InvertedGPSTransform
    A = 290.0
    B = 11400.0
    C = 11500.0


# -- gwpy.plotter.tex module tests --------------------------------------------

class TexTestCase(unittest.TestCase):
    def test_float_to_latex(self):
        self.assertEqual(float_to_latex(1), '1')
        self.assertEqual(float_to_latex(100), '10^{2}')
        self.assertEqual(float_to_latex(-500), r'-5\!\!\times\!\!10^{2}')
        self.assertRaises(TypeError, float_to_latex, '1')

    def test_label_to_latex(self):
        self.assertEqual(label_to_latex('Test'), 'Test')
        self.assertEqual(label_to_latex('Test_with_underscore'),
                         r'Test\_with\_underscore')

    def test_unit_to_latex(self):
        t = unit_to_latex(units.Hertz)
        self.assertEqual(t, r'$\mathrm{Hz}$')
        t = unit_to_latex(units.Volt.decompose())
        self.assertEqual(t, r'$\mathrm{m^{2}\,kg\,A^{-1}\,s^{-3}}$')


# -- gwpy.plotter.html module tests -------------------------------------------

class HtmlTestCase(unittest.TestCase):
    def setUp(self):
        self.data = numpy.vstack((numpy.arange(100), numpy.random.random(100)))

    def test_map_data(self):
        fig = figure()
        ax = fig.gca()
        html = map_data(self.data, ax, 'test.png')
        self.assertTrue(html.startswith('<!doctype html>'))
        html = map_data(self.data, ax, 'test.png', standalone=False)
        self.assertTrue(html.startswith('\n<img src="test.png"'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for table module
"""

import os.path
import tempfile
import unittest

from gwpy import version
from gwpy.table import lsctables

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

SNGL_BURST_XML_FILE = os.path.join(
    os.path.split(__file__)[0], 'data', 'H1-LDAS_STRAIN-968654552-10.xml.gz')


class TableTestMixin(object):
    TABLE_CLASS = None

    def test_has_gwpy_methods(self):
        for method in ['read', 'plot']:
            self.assertTrue(hasattr(self.TABLE_CLASS, method))


class SnglBurstTableTestCase(TableTestMixin, unittest.TestCase):
    """`TestCase` for `SnglBurstTable`
    """
    TABLE_CLASS = lsctables.SnglBurstTable

    def test_read_ligolw(self):
        table = self.TABLE_CLASS.read(SNGL_BURST_XML_FILE)
        table = self.TABLE_CLASS.read(SNGL_BURST_XML_FILE, format='ligolw')
        table = self.TABLE_CLASS.read(SNGL_BURST_XML_FILE, format='sngl_burst')
        self.assertEquals(len(table), 2052)

    def test_write_ligolw(self):
        # read table
        table = self.TABLE_CLASS.read(SNGL_BURST_XML_FILE)
        tmpxml = tempfile.mktemp(suffix='.xml.gz')
        # write table
        try:
            table.write(open(tmpxml, 'w'))
            table2 = self.TABLE_CLASS.read(tmpxml)
        finally:
            if os.path.isfile(tmpxml):
                os.remove(tmpxml)
        self.assertEquals(len(table), len(table2))
        self.assertEquals(table[0].get_peak(), table2[0].get_peak())
        self.assertEquals(table[0].snr, table2[0].snr)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for the time module
"""

import datetime

from compat import unittest

from astropy.units import (UnitConversionError, Quantity)

from gwpy import (version, time)

DATE = datetime.datetime(2000, 1, 1, 0, 0)
GPS = 630720013

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class TimeTests(unittest.TestCase):
    """`TestCase` for the time module
    """
    def test_to_gps(self):
        # test datetime conversion
        self.assertEqual(time.to_gps(DATE), GPS)
        # test Time
        self.assertEqual(
            time.to_gps(DATE, format='datetime', scale='utc'), GPS)
        # test tuple
        self.assertEqual(time.to_gps(tuple(DATE.timetuple())[:6]), GPS)
        # test Quantity
        self.assertEqual(time.to_gps(Quantity(GPS, 's')), GPS)
        # test errors
        self.assertRaises(UnitConversionError, time.to_gps, Quantity(1, 'm'))
        self.assertRaises(ValueError, time.to_gps, 'random string')

    def test_from_gps(self):
        date = time.from_gps(GPS)
        self.assertEqual(date, DATE)

    def test_tconvert(self):
        # from GPS
        date = time.tconvert(GPS)
        self.assertEqual(date, DATE)
        # from GPS using LAL LIGOTimeGPS
        try:
            from lal import LIGOTimeGPS
        except ImportError:
            pass
        else:
            d = time.tconvert(LIGOTimeGPS(GPS))
            self.assertEqual(d, DATE)
        # to GPS
        gps = time.tconvert(date)
        self.assertEqual(gps, GPS)
        # special cases
        now = time.tconvert()
        now2 = time.tconvert('now')
        self.assertEqual(now, now2)
        today = time.tconvert('today')
        yesterday = time.tconvert('yesterday')
        self.assertAlmostEqual(today - yesterday, 86400)
        self.assertTrue(now >= today)
        tomorrow = time.tconvert('tomorrow')
        self.assertAlmostEqual(tomorrow - today, 86400)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plot data from a LIGO_LW-format XML event table.

The :mod:`glue.ligolw` library provides a complete set of tools to read,
write, and manipulate data tables using the LIGO_LW XML schema, the
standard for storing transient event triggers generated from many
gravitational-wave search algorithms.

This modules provides a set of Axes and a Plot wrapper in order to
display these tables in x-y format, with optional colouring.
"""

from __future__ import division

import re
from six import string_types

import numpy
from matplotlib import (cm, collections, pyplot)
from matplotlib.projections import register_projection

from glue.ligolw.table import Table

from ..time import LIGOTimeGPS
from .core import Plot
from .timeseries import (TimeSeriesAxes, TimeSeriesPlot)
from .spectrum import SpectrumPlot
from .utils import float_to_latex
from ..table.utils import (get_table_column, get_row_value)

__all__ = ['EventTableAxes', 'EventTablePlot']


class EventTableAxes(TimeSeriesAxes):
    """Custom `Axes` for an :class:`~gwpy.plotter.EventTablePlot`.

    The `EventTableAxes` inherit from
    :class:`~gwpy.plotter.TimeSeriesAxes` as a convenience to
    optionally displaying a time-column. That choice has no effect on the
    rest of the `Axes` functionality.
    """
    name = 'triggers'

    def __init__(self, fig, *args, **kwargs):
        tsp = isinstance(fig, TimeSeriesPlot)
        if not tsp:
            kwargs.setdefault('xscale', 'linear')
        super(EventTableAxes, self).__init__(fig, *args, **kwargs)
        if not tsp:
            self.set_xlabel('')
            self.fmt_xdata = None

    def plot(self, *args, **kwargs):
        """Plot data onto these axes

        Parameters
        ----------
        *args
            a single :class:`~glue.ligolw.table.Table` (or sub-class)
            or anything valid for
            :meth:`~gwpy.plotter.TimeSeriesPlot.plot`.
        **kwargs
            keyword arguments applicable to
            :meth:`~matplotlib.axes.Axes.plot`
        """
        if isinstance(args[0], Table):
            return self.plot_table(*args, **kwargs)
        else:
            return super(EventTableAxes, self).plot(*args, **kwargs)

    def plot_table(self, table, x, y, color=None, size_by=None,
                   size_by_log=None, size_range=None, **kwargs):
        """Plot a LIGO_LW-format event `Table` onto these `Axes`

        Parameters
        ----------
        table : :class:`~glue.ligolw.table.Table`
            LIGO_LW-format XML event `Table` to display
        x : `str`
            name of column to display on the X-axis
        y : `str`
            name of column to display on the Y-axis
        c : `str`, optional
            name of column by which to colour the data
        **kwargs
            any other arguments applicable to
            :meth:`~matplotlib.axes.Axes.scatter`

        Returns
        -------
        collection
        """
        if size_by is not None and size_by_log is not None:
            raise ValueError("size_by_color and size_by_log_color are "
                             "mutually exclusive options, please select one")
        # get x-y data
        xdata = get_table_column(table, x)
        ydata = get_table_column(table, y)

        # rank data by size or colour
        sizecol = size_by or size_by_log or (size_range and color)
        if color:
            cdata = get_table_column(table, color)
        if sizecol:
            sdata = get_table_column(table, sizecol)
        if color and sizecol:
            zipped = zip(xdata, ydata, cdata, sdata)
            zipped.sort(key=lambda row: row[2])
            try:
                xdata, ydata, cdata, sdata = map(numpy.asarray, zip(*zipped))
            except ValueError:
                pass
        elif sizecol:
            zipped = zip(xdata, ydata, sdata)
            zipped.sort(key=lambda row: row[-1])
            try:
                xdata, ydata, sdata = map(numpy.asarray, zip(*zipped))
            except ValueError:
                pass
        elif color:
            zipped = zip(xdata, ydata, cdata)
            zipped.sort(key=lambda row: row[-1])
            try:
                xdata, ydata, cdata = map(numpy.asarray, zip(*zipped))
            except ValueError:
                pass

        # work out sizing
        if sizecol:
            if size_range is None and sdata.size:
                size_range = [sdata.min(), sdata.max()]
            if size_range:
                # convert color value into a size between the given min and max
                s = kwargs.pop('s', 20)
                sizes = [s/10., s]
                sp = (sdata - size_range[0]) / (size_range[1] - size_range[0])
                sp[sp < 0] = 0
                sp[sp > 1] = 1
                if size_by_log is None:
                    sarray = sizes[0] + sp * (sizes[1] - sizes[0])
                else:
                    logsizes = numpy.log10(sizes)
                    sarray = 10 ** (
                        logsizes[0] + sp * (logsizes[1] - logsizes[0]))
                kwargs.setdefault('s', sarray)

        if color:
            return self.scatter(xdata, ydata, c=cdata, **kwargs)
        else:
            return self.scatter(xdata, ydata, **kwargs)

    def plot_tiles(self, table, x, y, width, height, color=None,
                   anchor='center', edgecolors='face', linewidth=0.8,
                   **kwargs):
        # get x/y data
        xdata = get_table_column(table, x)
        ydata = get_table_column(table, y)
        wdata = get_table_column(table, width)
        hdata = get_table_column(table, height)

        # get color and sort
        if color:
            cdata = get_table_column(table, color)
            zipped = zip(xdata, ydata, wdata, hdata, cdata)
            zipped.sort(key=lambda row: row[-1])
            try:
                xdata, ydata, wdata, hdata, cdata = map(numpy.asarray,
                                                        zip(*zipped))
            except ValueError:
                pass

        # construct vertices
        if anchor == 'll':
            verts = [((x, y), (x, y+height), (x+width, y+height),
                      (x+width, y)) for (x, y, width, height) in
                     zip(xdata, ydata, wdata, hdata)]
        elif anchor == 'lr':
            verts = [((x-width, y), (x-width, y+height), (x, y+height),
                      (x, y)) for (x, y, width, height) in
                     zip(xdata, ydata, wdata, hdata)]
        elif anchor == 'ul':
            verts = [((x, y-height), (x, y), (x+width, y),
                      (x+width, y-height)) for (x, y, width, height) in
                     zip(xdata, ydata, wdata, hdata)]
        elif anchor == 'ur':
            verts = [((x-width, y-height), (x-width, y), (x, y),
                      (x, y-height)) for (x, y, width, height) in
                     zip(xdata, ydata, wdata, hdata)]
        elif anchor == 'center':
            verts = [((x-width/2., y-height/2.), (x-width/2., y+height/2.),
                      (x+width/2., y+height/2.), (x+width/2., y-height/2.))
                     for (x, y, width, height) in
                     zip(xdata, ydata, wdata, hdata)]
        else:
            raise ValueError("Unrecognised tile anchor '%s'." % anchor)

        # build collection
        cmap = kwargs.pop('cmap', pyplot.rcParams['image.cmap'])
        coll = collections.PolyCollection(verts, edgecolors=edgecolors,
                                          linewidth=linewidth, **kwargs)
        if color:
            coll.set_array(cdata)
            coll.set_cmap(cmap)

        return self.add_collection(coll)

    def add_loudest(self, table, rank, x, y, *columns, **kwargs):
        """Display the loudest event according to some rank.

        The loudest event is displayed as a gold star at its
        position given by the values in columns ``x``, and ``y``,
        and those values are displayed in a text box.

        Parameters
        ----------
        table : `~glue.ligolw.table.Table`
            LIGO_LW-format XML event table in which to find the loudest
            event
        rank : `str`
            name of column to use for ranking
        x : `str`
            name of column to display on the X-axis
        y : `str`
            name of column to display on the Y-axis
        color : `str`, optional
            name of column by which to colour the data
        **kwargs
            any other arguments applicable to
            :meth:`~matplotlib.axes.Axes.text`

        Returns
        -------
        out : `tuple`
            (`collection`, `text`) tuple of items added to the `Axes`
        """
        ylim = self.get_ylim()
        row = table[get_table_column(table, rank).argmax()]
        disp = "Loudest event:"
        columns = [x, y, rank] + list(columns)
        scat = []
        for i, column in enumerate(columns):
            if not column or column in columns[:i]:
                continue
            if i:
                disp += ','
            val = get_row_value(row, column)
            if i < 2:
                scat.append([float(val)])
            column = get_column_string(column)
            if pyplot.rcParams['text.usetex'] and column.endswith('Time'):
                disp += (r" %s$= %s$" % (column, LIGOTimeGPS(val)))
            elif pyplot.rcParams['text.usetex']:
                disp += (r" %s$=$ %s" % (column, float_to_latex(val, '%.3g')))
            else:
                disp += " %s = %.2g" % (column, val)
        disp = disp.rstrip(',')
        pos = kwargs.pop('position', [0.5, 1.00])
        kwargs.setdefault('transform', self.axes.transAxes)
        kwargs.setdefault('verticalalignment', 'bottom')
        kwargs.setdefault('horizontalalignment', 'center')
        args = pos + [disp]
        self.scatter(*scat, marker='*', zorder=1000, facecolor='gold',
                     edgecolor='black',  s=200)
        self.text(*args, **kwargs)
        if self.get_title():
            pos = self.title.get_position()
            self.title.set_position((pos[0], pos[1] + 0.05))
        self.set_ylim(*ylim)

register_projection(EventTableAxes)


class _EventTableMetaPlot(type):
    """Meta-class for generating a new :class:`EventTablePlot`.

    This object allows the choice of parent class for the
    `EventTablePlot` to be made at runtime, dependent on the given
    x-column of the first displayed Table.
    """
    def __call__(cls, *args, **kwargs):
        """Execute the meta-class, given the arguments for the plot

        All ``*args`` and ``**kwargs`` are those passed to the
        `EventTablePlot` constructor, used to determine the appropriate
        parent class the that object at runtime.
        """
        # find x-column: copy the arguments and find the strings
        a2 = list(args)
        while len(a2):
            if isinstance(a2[0], string_types):
                break
            a2.pop(0)
        # if at least one string was found, treat it as the x-axis column name
        if 'base' in kwargs:
            plotclass = kwargs.pop('base')
        elif len(a2):
            xcol = a2[0]
            # initialise figure as a TimeSeriesPlot
            if re.search('time\Z', xcol, re.I):
                plotclass = TimeSeriesPlot
            # or as a SpectrumPlot
            elif re.search('(freq\Z|frequency\Z)', xcol, re.I):
                plotclass = SpectrumPlot
            # otherwise as a standard Plot
            else:
                plotclass = Plot
        else:
            plotclass = Plot
        cls.__bases__ = (plotclass,)
        return super(_EventTableMetaPlot, cls).__call__(*args, **kwargs)


class EventTablePlot(TimeSeriesPlot, SpectrumPlot, Plot):
    """`Figure` for displaying a :class:`~glue.ligolw.table.Table`.

    Parameters
    ----------
    table : :class:`~glue.ligolw.table.Table`
        LIGO_LW-format XML event `Table` to display
    x : `str`
        name of column to display on the X-axis
    y : `str`
        name of column to display on the Y-axis
    c : `str`, optional
        name of column by which to colour the data
    **kwargs
        any other arguments applicable to the `Plot` constructor, and
        the `Table` plotter.

    Returns
    -------
    plot : :class:`EventTablePlot`
        new plot for displaying tabular data.

    Notes
    -----
    The form of the returned `EventTablePlot` is decided at run-time,
    rather than when the module was imported.
    If tables are passed directly to the constructor, for example::

        >>> plot = EventTablePlot(table1, 'time', 'snr')

    the columns as given are used to determine the appropriate parent
    class for the output.

    If the input x-column (the first string argument) ends with 'time'
    the output is a child of the :class:`~gwpy.plotter.TimeSeriesPlot`,
    allowing easy formatting of GPS times, while if the x-column ends with
    'frequency', the output comes from the
    :class:`~gwpy.plotter.SpectrumPlot`, otherwise the parent is
    the core :class:`~gwpy.plotter.Plot`.
    """
    _DefaultAxesClass = EventTableAxes
    __metaclass__ = _EventTableMetaPlot

    def __init__(self, *args, **kwargs):
        """Generate a new `EventTablePlot`.
        """
        # extract plotting keyword arguments
        plotargs = dict()
        for arg in ['linewidth', 'facecolor', 'edgecolor', 'marker', 'cmap',
                    's', 'size_by', 'size_by_log', 'size_range', 'label',
                    'edgecolors', 'facecolors', 'linewidths', 'antialiaseds',
                    'offsets', 'transOffset', 'norm', 'rasterized']:
            if arg in kwargs:
                val = kwargs.pop(arg, None)
                if val is not None:
                    plotargs[arg] = val

        # extract columns
        args = list(args)
        tables = []
        columns = {}
        tiles = False
        while len(args):
            arg = args[0]
            if isinstance(arg, string_types):
                break
            tables.append(args.pop(0))
        if len(tables) != 0 and len(args) < 2:
            raise ValueError("columnnames for 'x' and 'y' axes must be given "
                             "after any tables, e.g. "
                             "TablePlot(t1, t2, 'time', 'snr')")
        if len(args) in [3, 5]:
            kwargs.setdefault('color', args.pop(-1))
        elif len(args) == 4:
            tiles = True
        elif len(args) > 5:
            raise ValueError("No more than three column names should be given")
        if len(tables):
            columns = dict(zip(['x', 'y', 'width', 'height'], args))

        # extract column arguments
        epoch = kwargs.pop('epoch', None)
        sep = kwargs.pop('sep', False)
        columns['color'] = kwargs.pop('color', None)
        if columns['color'] and len(tables) > 1 and not sep:
            raise ValueError("Plotting multiple Tables on a single set of "
                             "Axes with a colorbar is not supported, "
                             "currently...")

        super(EventTablePlot, self).__init__(**kwargs)

        # plot data
        for table in tables:
            if len(args) == 2:
                self.add_table(table, columns['x'], columns['y'],
                               color=columns['color'], newax=sep, **plotargs)
            elif len(args) == 4:
                self.add_tiles(table, columns['x'], columns['y'],
                               columns['width'], columns['height'],
                               color=columns['color'], newax=sep, **plotargs)
        if len(tables):
            # set auto-scale
            for ax in self.axes:
                ax.autoscale(axis='both', tight=True)
            # set individual epoch for TimeSeriesAxes
            if isinstance(self, TimeSeriesPlot) and sep:
                for ax, table in zip(self.axes, tables):
                    axepoch = epoch
                    if axepoch is None:
                        tcol = numpy.asarray(get_table_column(table,
                                                              columns['x']))
                        if epoch is None:
                            epoch = tcol.min()
                        else:
                            epoch = min(epoch, tcol.min())
                    ax.set_epoch(epoch)
            # set global epoch for TimeSeriesPlot
            elif isinstance(self, TimeSeriesPlot):
                ax = self.axes[0]
                if epoch is None:
                    for table in tables:
                        tcol = numpy.asarray(get_table_column(table,
                                                              columns['x']))
                        if epoch is None:
                            epoch = tcol.min()
                        else:
                            epoch = min(epoch, tcol.min())
                ax.set_epoch(epoch)
            # remove labels
            if sep:
                for ax in self.axes[:-1]:
                    ax.set_xlabel("")

    def add_table(self, table, x, y, color=None, projection='triggers',
                  ax=None, newax=None, **kwargs):
        """Add a LIGO_LW Table to this Plot

        Parameters
        ----------
        table : :class:`~glue.ligolw.table.Table`
            LIGO_LW-format XML event `Table` to display
        x : `str`
            name of column to display on the X-axis
        y : `str`
            name of column to display on the Y-axis
        c : `str`, optional
            name of column by which to colour the data
        projection : `str`, optiona, default: ``'triggers'``
            name of the Axes projection on which to plot data
        ax : :class:`~gwpy.plotter.Axes`, optional
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs.
            other keyword arguments passed to the
            :meth:`EventTableAxes.plot_table` method

        Returns
        -------
        scatter : :class:`~matplotlib.collections.Collection`
            the displayed collection for this `Table`

        See Also
        --------
        :meth:`EventTableAxes.plot_table`
            for details on arguments and keyword arguments other than
            ``ax`` and ``newax`` for this method.
        """
        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection)
        return ax.plot_table(table, x, y, color=color, **kwargs)

    def add_tiles(self, table, x, y, width, height, color=None,
                  anchor='center', projection='triggers', ax=None,
                  newax=None, **kwargs):
        """Add a LIGO_LW Table to this Plot

        Parameters
        ----------
        table : :class:`~glue.ligolw.table.Table`
            LIGO_LW-format XML event `Table` to display
        x : `str`
            name of column for tile x-anchor
        y : `str`
            name of column for tile y-anchor
        width : `str`
            name of column for tile width
        height : `str`
            name of column for tile height
        color : `str`, optional
            name of column by which to colour the data
        anchor : `str`, optional, default: ``'center'``
            position of (x, y) vertex on tile, default 'center'.
            Other options: 'll', 'lr', 'ul', 'ur'.
        projection : `str`, optiona, default: ``'triggers'``
            name of the Axes projection on which to plot data
        ax : :class:`~gwpy.plotter.Axes`, optional
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs.
            other keyword arguments passed to the
            :meth:`EventTableAxes.plot_table` method

        Returns
        -------
        scatter : :class:`~matplotlib.collections.Collection`
            the displayed collection for this `Table`

        See Also
        --------
        :meth:`EventTableAxes.plot_table`
            for details on arguments and keyword arguments other than
            ``ax`` and ``newax`` for this method.
        """
        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection)
        return ax.plot_tiles(table, x, y, width, height, color=color,
                             anchor=anchor, **kwargs)


def get_column_string(column):
    """
    Format the string columnName (e.g. xml table column) into latex format for.
    an axis label.

    Examples:

    >>> get_column_string('snr')
    'SNR'

    >>> get_column_string('bank_chisq_dof')
    r'Bank $\chi^2$ DOF'

    Arguments:

      columnName : string
        string to format
    """
    acro = ['snr', 'ra', 'dof', 'id', 'ms', 'far']
    greek = ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'eta',
             'theta', 'iota', 'kappa', 'lamda', 'mu', 'nu', 'xi', 'omicron',
             'pi', 'rho', 'sigma', 'tau', 'upsilon', 'phi', 'chi', 'psi',
             'omega']
    unit = ['ns']
    sub = ['flow', 'fhigh', 'hrss', 'mtotal', 'mchirp']

    tex = pyplot.rcParams['text.usetex']

    words = []
    for w in re.split('\s', column):
        if w.isupper():
            words.append(w)
        else:
            words.extend(re.split('_', w))

    # parse words
    for i, w in enumerate(words):
        # get acronym in lower case
        if w in acro:
            words[i] = w.upper()
        # get numerical unit
        elif w in unit:
            words[i] = '$(%s)$' % w
        # get character with subscript text
        elif w in sub and tex:
            words[i] = '%s$_{\mbox{\\small %s}}$' % (w[0], w[1:])
        # get greek word
        elif w in greek and tex:
            words[i] = '$\%s$' % w
        # get starting with greek word
        elif re.match('(%s)' % '|'.join(greek), w) and tex:
            if w[-1].isdigit():
                words[i] = '$\%s_{%s}$''' % tuple(
                    re.findall(r"[a-zA-Z]+|\d+", w))
            elif w.endswith('sq'):
                words[i] = '$\%s^2$' % w.rstrip('sq')
        # get everything else
        else:
            if w.isupper():
                words[i] = w
            else:
                words[i] = w.title()
            # escape underscore
            words[i] = re.sub(r'(?<!\\)_', r'\_', words[i])
    return ' '.join(words)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for data classes
"""

import abc
import os
import tempfile

from compat import unittest

from numpy import testing as nptest
import numpy

from astropy import units
from astropy.time import Time

from gwpy import version
from gwpy.data import (Array, Series)
from gwpy.detector import Channel

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

SEED = 1
GPS_EPOCH = 12345
TIME_EPOCH = Time(12345, format='gps', scale='utc')
CHANNEL_NAME = 'X1:TEST-CHANNEL'
CHANNEL = Channel(CHANNEL_NAME)


class CommonTests(object):
    __metaclass_ = abc.ABCMeta
    TEST_CLASS = Array
    tmpfile = '%s.%%s' % tempfile.mktemp(prefix='gwpy_test_')

    def setUp(self, dtype=None):
        numpy.random.seed(SEED)
        self.data = (numpy.random.random(100) * 1e5).astype(dtype=dtype)
        self.datasq = self.data ** 2

    @property
    def TEST_ARRAY(self):
        try:
            return self._TEST_ARRAY
        except AttributeError:
            self._TEST_ARRAY = self.create(name='test', unit='meter',
                                           channel=CHANNEL_NAME,
                                           epoch=TIME_EPOCH)
            return self.TEST_ARRAY

    def create(self, *args, **kwargs):
        kwargs.setdefault('copy', False)
        return self.TEST_CLASS(self.data, *args, **kwargs)

    def assertArraysEqual(self, ts1, ts2, *args):
        nptest.assert_array_equal(ts1.value, ts2.value)
        if not args:
            args = ['units'] + self.TEST_CLASS._metadata_slots
        for attr in args:
            self.assertEqual(getattr(ts1, attr, None),
                             getattr(ts2, attr, None))

    # -- test methods ---------------------------

    def test_init(self):
        """Test Array creation
        """
        # test basic empty contructor
        self.assertRaises(TypeError, self.TEST_CLASS)
        self.assertRaises(IndexError, self.TEST_CLASS, [])
        # test with some data
        array = self.create()
        nptest.assert_array_equal(array.value, self.data)

    def test_unit(self):
        array = self.create()
        self.assertIsNone(array.unit)
        array = self.create(unit='m')
        self.assertEquals(array.unit, units.m)

    def test_name(self):
        array = self.create(name=None)
        self.assertIsNone(array.name)
        array = self.create(name='TEST CASE')
        self.assertEquals(array.name, 'TEST CASE')

    def test_epoch(self):
        array = self.create()
        self.assertIsNone(array.epoch)
        array = self.create(epoch=GPS_EPOCH)
        self.assertEquals(array.epoch, TIME_EPOCH)

    def test_channel(self):
        array = self.create()
        self.assertIsNone(array.channel)
        array = self.create(channel=CHANNEL_NAME)
        self.assertEquals(array.channel, CHANNEL)

    def test_math(self):
        """Test Array math operations
        """
        array = self.create(unit='Hz')
        # test basic operations
        arraysq = array ** 2
        nptest.assert_array_equal(arraysq.value, self.datasq)
        self.assertEqual(arraysq.unit, units.Hz ** 2)
        self.assertEqual(arraysq.name, array.name)
        self.assertEqual(arraysq.epoch, array.epoch)
        self.assertEqual(arraysq.channel, array.channel)

    def test_copy(self):
        """Test Array.copy
        """
        array = self.create(unit='Hz')
        array2 = array.copy()
        nptest.assert_array_equal(array.value, array2.value)
        self.assertEqual(array.unit, array2.unit)

    # -- test I/O -------------------------------

    def test_hdf5_write(self, delete=True, format=[None, 'hdf5', 'hdf']):
        if isinstance(format, (list, tuple)):
            formats = format
        else:
            formats = [format]
        hdfout = self.tmpfile % 'hdf'
        for format in formats:
            try:
                self.TEST_ARRAY.write(hdfout, format=format)
            except ImportError as e:
                self.skipTest(str(e))
            finally:
                if delete and os.path.isfile(hdfout):
                    os.remove(hdfout)
        return hdfout

    def test_hdf5_read(self, format=[None, 'hdf5', 'hdf']):
        try:
            hdfout = self.test_hdf5_write(delete=False, format='hdf5')
        except ImportError as e:
            self.skipTest(str(e))
        else:
            if isinstance(format, (list, tuple)):
                formats = format
            else:
                formats = [format]
            try:
                for format in formats:
                    ts = self.TEST_CLASS.read(hdfout, format=format)
                    self.assertArraysEqual(self.TEST_ARRAY, ts)
            finally:
                if os.path.isfile(hdfout):
                    os.remove(hdfout)


class ArrayTestCase(CommonTests, unittest.TestCase):
    pass


class SeriesTestCase(CommonTests, unittest.TestCase):
    TEST_CLASS = Series

    def test_init(self):
        super(SeriesTestCase, self).test_init()
        series = self.create(x0=0, dx=1)
        self.assertEqual(series.x0, units.Quantity(0, series._default_xunit))
        self.assertEqual(series.dx, units.Quantity(1, series._default_xunit))

    def test_xunit(self, unit=None):
        if unit is None:
            unit = self.TEST_CLASS._default_xunit
        series = self.create(unit='Hz', dx=4*unit)
        self.assertEqual(series.x0, 0*unit)
        self.assertEqual(series.dx, 4*unit)
        # for series only, test arbitrary xunit
        if self.TEST_CLASS == Series:
            series = self.create(unit='Hz', dx=4*units.m)
            self.assertEqual(series.x0, 0*units.m)
            self.assertEqual(series.dx, 4*units.m)

    def test_index(self):
        series = self.create()
        self.assertFalse(hasattr(series, '_xindex'))
        nptest.assert_array_equal(
            series.xindex, numpy.arange(series.size) * series.dx + series.x0)

    def test_pickle(self):
        """Check pickle-unpickle yields unchanged data
        """
        import cPickle
        ts = self.create()
        pickle = ts.dumps()
        ts2 = cPickle.loads(pickle)
        self.assertArraysEqual(ts, ts2)

    def test_crop(self):
        """Test cropping `Series` by GPS times
        """
        ts = self.create()
        ts2 = ts.crop(10, 20)
        self.assertEqual(ts2.x0.value, 10)
        self.assertEqual(ts2.xspan[1], 20)
        nptest.assert_array_equal(ts2.value, ts.value[10:20])

    def test_is_compatible(self):
        """Test the `Series.is_compatible` method
        """
        ts1 = self.create()
        ts2 = self.create(name='TEST CASE 2')
        self.assertTrue(ts1.is_compatible(ts2))
        ts3 = self.create(dx=2)
        self.assertRaises(ValueError, ts1.is_compatible, ts3)
        ts4 = self.create(unit='m')
        self.assertRaises(ValueError, ts1.is_compatible, ts4)

    def test_is_contiguous(self):
        """Test the `Series.is_contiguous` method
        """
        ts1 = self.create()
        ts2 = self.create(x0=ts1.xspan[1])
        self.assertEquals(ts1.is_contiguous(ts2), 1)
        self.assertEquals(ts1.is_contiguous(ts2.value), 1)
        ts3 = self.create(x0=ts1.xspan[1]+1)
        self.assertEquals(ts1.is_contiguous(ts3), 0)
        ts4 = self.create(x0=-ts1.xspan[1])
        self.assertEquals(ts1.is_contiguous(ts4), -1)

    def test_append(self):
        """Test the `Series.append` method
        """
        # create arrays
        ts1 = self.create()
        ts2 = self.create(x0=ts1.xspan[1])
        # test basic append
        ts3 = ts1.append(ts2, inplace=False)
        self.assertEquals(ts3.epoch, ts1.epoch)
        self.assertEquals(ts3.x0, ts1.x0)
        self.assertEquals(ts3.size, ts1.size+ts2.size)
        self.assertEquals(ts3.xspan, ts1.xspan+ts2.xspan)
        self.assertRaises(ValueError, ts3.append, ts1)
        nptest.assert_array_equal(ts3.value[:ts1.size], ts1.value)
        nptest.assert_array_equal(ts3.value[-ts2.size:], ts2.value)
        # test appending with one xindex deletes it in the output
        ts1.xindex
        ts4 = ts1.append(ts2, inplace=False)
        self.assertTrue(hasattr(ts4, '_xindex'))
        nptest.assert_array_equal(
            ts4.xindex.value,
            numpy.concatenate((ts1.xindex.value, ts2.xindex.value)))
        # test appending with both xindex appends as well
        ts1.xindex
        ts2.xindex
        ts5 = ts1.append(ts2, inplace=False)
        self.assertTrue(hasattr(ts5, '_xindex'))
        nptest.assert_array_equal(
            ts5.xindex.value,
            numpy.concatenate((ts1.xindex.value, ts2.xindex.value)))

    def test_prepend(self):
        """Test the `Series.prepend` method
        """
        ts1 = self.create()
        ts2 = self.create(x0=ts1.xspan[1]) * 2
        ts3 = ts2.prepend(ts1, inplace=False)
        self.assertEquals(ts3.x0, ts1.x0)
        self.assertEquals(ts3.size, ts1.size+ts2.size)
        self.assertEquals(ts3.xspan, ts1.xspan+ts2.xspan)
        self.assertRaises(ValueError, ts3.prepend, ts1)
        nptest.assert_array_equal(ts3.value[:ts1.size], ts1.value)
        nptest.assert_array_equal(ts3.value[-ts2.size:], ts2.value)

    def test_update(self):
        """Test the `Series.update` method
        """
        ts1 = self.create()
        ts2 = self.create(x0=ts1.xspan[1])[:ts1.size//2]
        ts3 = ts1.update(ts2, inplace=False)
        self.assertEquals(ts3.x0, ts1.x0 + abs(ts2.xspan)*ts1.x0.unit)
        self.assertEquals(ts3.size, ts1.size)
        self.assertRaises(ValueError, ts3.update, ts1)

    def test_pad(self):
        """Test the `Series.pad` method
        """
        ts1 = self.create()
        ts2 = ts1.pad(10)
        self.assertEquals(ts2.size, ts1.size + 20)
        nptest.assert_array_equal(
            ts2.value,
            numpy.concatenate((numpy.zeros(10), ts1.value, numpy.zeros(10))))
        self.assertEquals(ts2.x0, ts1.x0 - 10*ts1.x0.unit)
        # test pre-pad
        ts3 = ts1.pad((20, 10))
        self.assertEquals(ts3.size, ts1.size + 30)
        nptest.assert_array_equal(
            ts3.value,
            numpy.concatenate((numpy.zeros(20), ts1.value, numpy.zeros(10))))
        self.assertEquals(ts3.x0, ts1.x0 - 20*ts1.x0.unit)
        # test bogus input
        self.assertRaises(ValueError, ts1.pad, -1)

    def test_diff(self):
        """Test the `Series.diff` method

        This just ensures that the returned `Series` has the right length
        and the right x0
        """
        ts1 = self.create()
        diff = ts1.diff()
        self.assertIsInstance(diff, type(ts1))
        self.assertEqual(ts1.size - 1, diff.size)
        self.assertEqual(diff.x0, ts1.x0 + ts1.dx)
        self.assertEqual(diff.xspan[1], ts1.xspan[1])
        self.assertEqual(diff.channel, ts1.channel)
        # test n=3
        diff = ts1.diff(n=3)
        self.assertEqual(ts1.size - 3, diff.size)
        self.assertEqual(diff.x0, ts1.x0 + ts1.dx * 3)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for segments module
"""

import os.path
import tempfile
import StringIO
from six import PY3
from urllib2 import (urlopen, URLError)

from compat import unittest

from glue.segments import PosInfinity

from gwpy import version
from gwpy.segments import (Segment, SegmentList,
                           DataQualityFlag, DataQualityDict)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

KNOWN = SegmentList([
    Segment(0, 3),
    Segment(6, 7),
])
ACTIVE = SegmentList([
    Segment(1, 2),
    Segment(3, 4),
    Segment(5, 7),
])
KNOWNACTIVE = SegmentList([
    Segment(1, 2),
    Segment(6, 7),
])

KNOWN2 = SegmentList([Segment(100, 150)])
ACTIVE2 = SegmentList([
    Segment(100, 101),
    Segment(110, 120),
])

# get padding stuff
PADDING = (-0.5, 1)
KNOWNPAD = SegmentList([
    Segment(-.5, 4),
    Segment(5.5, 8),
])
ACTIVEPAD = SegmentList([
    Segment(.5, 3),
    Segment(2.5, 5),
    Segment(4.5, 8),
])
ACTIVEPADC = SegmentList([
    Segment(.5, 4),
    Segment(5.5, 8),
])

SEGXML = os.path.join(os.path.split(__file__)[0], 'data',
                      'X1-GWPY_TEST_SEGMENTS-0-10.xml.gz')
SEGWIZ = os.path.join(os.path.split(__file__)[0], 'data',
                      'X1-GWPY_TEST_SEGMENTS-0-10.txt')
FLAG1 = 'X1:GWPY-TEST_SEGMENTS:1'
FLAG2 = 'X1:GWPY-TEST_SEGMENTS:2'

QUERY_START = 1108598416
QUERY_END = 1108684816
QUERY_KNOWN = SegmentList([(1108598416, 1108632895), (1108632901, 1108684816)])
QUERY_ACTIVE = SegmentList([(1108623497, 1108624217)])
QUERY_FLAG = 'L1:DMT-DC_READOUT:1'
QUERY_URL = 'https://segments.ligo.org'

VETO_DEFINER_FILE = ('https://www.lsc-group.phys.uwm.edu/ligovirgo/cbc/public/'
                     'segments/ER7/H1L1V1-ER7_CBC_OFFLINE.xml')
ER7_START = 'June 3'
ER7_END = 'June 14'


class SegmentListTests(unittest.TestCase):
    """Unit tests for the `SegmentList` class
    """
    tmpfile = '%s.%%s' % tempfile.mktemp(prefix='gwpy_test_segmentlist')

    def test_read_segwizard(self):
        active = SegmentList.read(SEGWIZ, coalesce=False)
        self.assertTrue(active == ACTIVE,
                        'SegmentList.read(segwizard) mismatch:\n\n%s\n\n%s'
                        % (ACTIVE, active))

    def test_write_segwizard(self):
        tmpfile = self.tmpfile % 'txt'
        ACTIVE.write(tmpfile)
        self.assertTrue(open(tmpfile, 'r').read() == open(SEGWIZ, 'r').read(),
                        'SegmentList.write(segwizard) mismatch, %s '
                        'differs from %s' % (tmpfile, SEGWIZ))
        os.remove(tmpfile)


class DataQualityFlagTests(unittest.TestCase):
    """Unit tests for the `DataQualityFlag` class
    """
    tmpfile = '%s.%%s' % tempfile.mktemp(prefix='gwpy_test_dqflag')

    def test_coalesce(self):
        flag = DataQualityFlag(FLAG1, active=ACTIVE, known=KNOWN)
        self.assertFalse(flag.regular,
                         'flag.regular test failed (should be False)')
        flag.coalesce()
        self.assertTrue(flag.known == KNOWN, 'flag.known changed by coalesce')
        self.assertTrue(flag.active == KNOWNACTIVE,
                        'flag.active misset by coalesce')
        self.assertTrue(flag.regular,
                        'flag.regular test failed (should be True)')

    def test_read_segwizard(self):
        flag = DataQualityFlag.read(SEGWIZ, FLAG1, coalesce=False)
        self.assertTrue(flag.active == ACTIVE,
                        'DataQualityFlag.read(segwizard) mismatch:\n\n%s\n\n%s'
                        % (ACTIVE, flag.active))
        self.assertTrue(flag.known == flag.active)

    def test_write_segwizard(self):
        tmpfile = self.tmpfile % 'txt'
        DataQualityFlag(FLAG1, active=ACTIVE).write(tmpfile)
        self.assertTrue(open(tmpfile, 'r').read() == open(SEGWIZ, 'r').read(),
                        'DataQualityFlag.write(segwizard) mismatch, %s '
                        'differs from %s' % (tmpfile, SEGWIZ))
        os.remove(tmpfile)

    def test_fail_write_segwizard(self):
        flag = DataQualityFlag(FLAG1, active=ACTIVE, known=KNOWN)
        self.assertRaises(ValueError, flag.write, StringIO.StringIO,
                          format='segwizard')

    def test_read_ligolw(self):
        flag = DataQualityFlag.read(SEGXML, FLAG1, coalesce=False)
        self.assertTrue(flag.active == ACTIVE,
                        'DataQualityFlag.read(ligol) mismatch:\n\n%s\n\n%s'
                        % (ACTIVE, flag.active))
        self.assertTrue(flag.known == KNOWN,
                        'DataQualityFlag.read(ligol) mismatch:\n\n%s\n\n%s'
                        % (KNOWN, flag.known))

    def test_write_ligolw(self):
        if PY3:
            types = [str]
        else:
            types = [str, unicode]
        for type_ in types:
            tmpfile = type_(self.tmpfile % 'xml.gz')
            try:
                DataQualityFlag(FLAG1, active=ACTIVE, known=KNOWN).write(tmpfile)
            finally:
                os.remove(tmpfile)

    def test_write_hdf5(self, delete=True):
        flag = DataQualityFlag(FLAG1, active=ACTIVE, known=KNOWN)
        hdfout = self.tmpfile % 'hdf'
        try:
            flag.write(hdfout)
        except ImportError as e:
            self.skipTest(str(e))
        else:
            if delete:
                os.remove(hdfout)
        return hdfout

    def test_read_hdf5(self):
        try:
            hdfout = self.test_write_hdf5(delete=False)
        except ImportError as e:
            self.skipTest(str(e))
        else:
            flag = DataQualityFlag.read(hdfout)
            os.remove(hdfout)
            self.assertTrue(flag.active == ACTIVE,
                            'DataQualityFlag.read(hdf5) mismatch:\n\n%s\n\n%s'
                            % (ACTIVE, flag.active))
            self.assertTrue(flag.known == KNOWN,
                            'DataQualityFlag.read(hdf5) mismatch:\n\n%s\n\n%s'
                            % (KNOWN, flag.known))

    def _query(self, cm, *args, **kwargs):
        try:
            return cm(*args, **kwargs)
        except (ImportError, URLError) as e:
            self.skipTest(str(e))
        except AttributeError as e:
            if 'PKCS5_SALT_LEN' in str(e):
                self.skipTest(str(e))
            else:
                raise

    def test_query(self):
        flag = self._query(DataQualityFlag.query,
                           QUERY_FLAG, QUERY_START, QUERY_END, url=QUERY_URL)
        self.assertEqual(flag.known, QUERY_KNOWN)
        self.assertEqual(flag.active, QUERY_ACTIVE)

    def test_query_dqsegdb(self):
        flag = self._query(DataQualityFlag.query_dqsegdb,
                           QUERY_FLAG, QUERY_START, QUERY_END, url=QUERY_URL)
        self.assertEqual(flag.known, QUERY_KNOWN)
        self.assertEqual(flag.active, QUERY_ACTIVE)

    def test_query_dqsegdb_versionless(self):
        flag = self._query(DataQualityFlag.query_dqsegdb,
                           QUERY_FLAG.rsplit(':', 1)[0], QUERY_START,
                           QUERY_END, url=QUERY_URL)
        self.assertEqual(flag.known, QUERY_KNOWN)
        self.assertEqual(flag.active, QUERY_ACTIVE)

    def test_query_dqsegdb_multi(self):
        querymid = int(QUERY_START + (QUERY_END - QUERY_START) /2.)
        segs = SegmentList([Segment(QUERY_START, querymid),
                            Segment(querymid, QUERY_END)])
        flag = self._query(DataQualityFlag.query_dqsegdb,
                           QUERY_FLAG, segs, url=QUERY_URL)
        self.assertEqual(flag.known, QUERY_KNOWN)
        self.assertEqual(flag.active, QUERY_ACTIVE)

    def test_pad(self):
        flag = DataQualityFlag(FLAG1, active=ACTIVE, known=KNOWN)
        # test without arguments (and no padding)
        padded = flag.pad()
        self.assertListEqual(padded.known, flag.known)
        self.assertListEqual(padded.active, flag.active)
        # test without arguments (and no padding)
        flag.padding = PADDING
        padded = flag.pad()
        self.assertListEqual(padded.known, KNOWNPAD)
        self.assertListEqual(padded.active, ACTIVEPAD)
        # test with arguments
        flag = DataQualityFlag(FLAG1, active=ACTIVE, known=KNOWN)
        padded = flag.pad(*PADDING)
        self.assertListEqual(padded.known, KNOWNPAD)
        self.assertListEqual(padded.active, ACTIVEPAD)
        # test coalesce
        padded.coalesce()
        self.assertListEqual(padded.active, ACTIVEPADC)


class DataQualityDictTestCase(unittest.TestCase):
    tmpfile = '%s.%%s' % tempfile.mktemp(prefix='gwpy_test_dqdict')
    VETO_DEFINER = tmpfile % 'vdf.xml'

    def setUp(self):
        # download veto definer
        vdffile = urlopen(VETO_DEFINER_FILE)
        with open(self.VETO_DEFINER, 'w') as f:
            f.write(vdffile.read())

    def tearDown(self):
        if os.path.isfile(self.VETO_DEFINER):
            os.remove(self.VETO_DEFINER)

    def test_from_veto_definer_file(self):
        vdf = DataQualityDict.from_veto_definer_file(self.VETO_DEFINER)
        self.assertNotEqual(len(vdf.keys()), 0)
        # test missing h(t) flag
        self.assertIn('H1:DCH-MISSING_H1_HOFT_C00:1', vdf)
        self.assertEquals(vdf['H1:DCH-MISSING_H1_HOFT_C00:1'].known[0][0],
                          1073779216)
        self.assertEquals(vdf['H1:DCH-MISSING_H1_HOFT_C00:1'].known[0][-1],
                          PosInfinity)
        self.assertEquals(vdf['H1:DCH-MISSING_H1_HOFT_C00:1'].category, 1)
        # test injections padding
        self.assertEquals(vdf['H1:ODC-INJECTION_CBC:1'].padding, Segment(-8, 8))

    def test_read_ligolw(self):
        flags = DataQualityDict.read(SEGXML)
        self.assertEquals(len(flags.keys()), 2)
        self.assertIn(FLAG1, flags)
        self.assertIn(FLAG2, flags)
        flags = DataQualityDict.read(SEGXML, [FLAG2])
        self.assertEquals(len(flags.keys()), 1)
        self.assertEquals(flags[FLAG2].known, KNOWN2)
        self.assertEquals(flags[FLAG2].active, ACTIVE2)

    def test_write_ligolw(self):
        tmpfile = self.tmpfile % 'xml.gz'
        try:
            flags = DataQualityDict.read(SEGXML)
        except Exception as e:
            self.skipTest(str(e))
        try:
            flags.write(tmpfile)
        finally:
            os.remove(tmpfile)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Series` is a one-dimensional array with metadata
"""

import numpy

from astropy.units import (Unit, Quantity)

from .series import Series
from .. import version
from ..utils.docstring import interpolate_docstring

__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

interpolate_docstring.update(
    ArrayYaxis=(
        """y0 : `float`, `~astropy.units.Quantity`, optional, default: `0`
        the starting value for the y-axis of this array

    dy : `float`, `~astropy.units.Quantity, optional, default: `1`
        the step size for the y-axis of this array

    yindex : `array-like`
        the complete array of y-axis values for this array. This argument
        takes precedence over `y0` and `dy` so should be
        given in place of these if relevant, not alongside"""),
)


@interpolate_docstring
class Array2D(Series):
    """A two-dimensional array with metadata

    Parameters
    ----------
    %(Array1)s

    %(ArrayXaxis)s

    %(ArrayYaxis)s

    %(Array2)s

    Returns
    -------
    array : `Array`
        a new array, with a view of the data, and all associated metadata
    """
    _metadata_slots = Series._metadata_slots + ['y0', 'dy', 'yindex']
    _default_xunit = Unit('')
    _default_yunit = Unit('')
    _ndim = 2

    def __new__(cls, data, unit=None, xindex=None, yindex=None, x0=0,
                dx=1, y0=0, dy=1, **kwargs):
        """Define a new `Array2D`
        """
        new = super(Array2D, cls).__new__(cls, data, unit=unit, xindex=xindex,
                                          x0=0, dx=dx, **kwargs)
        if isinstance(y0, Quantity):
            yunit = y0.unit
        elif isinstance(dy, Quantity):
            yunit = dy.unit
        else:
            yunit = cls._default_yunit
        if y0 is not None:
            new.y0 = Quantity(y0, yunit)
        if dy is not None:
            new.dy = Quantity(dy, yunit)
        if yindex is not None:
            new.yindex = yindex
        return new

    # rebuild getitem to handle complex slicing
    def __getitem__(self, item):
        new = super(Array2D, self).__getitem__(item)
        # unwrap item request
        if isinstance(item, tuple):
            x, y = item
        else:
            x = item
            y = None
        # extract a Quantity
        if isinstance(new, (float, int)):
            return Quantity(new, unit=self.unit)
        # unwrap a Series
        if len(new.shape) == 1:
            new = new.value.view(Series)
            if isinstance(x, (float, int)):
                new.dx = self.dy
                new.x0 = self.y0
        # unwrap a Spectrogram
        else:
            new = new.value.view(type(self))
            new.__dict__ = self.copy_metadata()
        # update metadata
        if isinstance(x, slice):
            if x.start:
                new.x0 = new.x0 + x.start * new.dx
            if x.step:
                new.dx = new.dx + x.step
        if len(new.shape) == 1 and isinstance(y, slice):
            if y.start:
                new.x0 = new.x0 + y.start * new.dx
            if y.step:
                new.dx = new.dx * y.step
        elif isinstance(y, slice):
            if y.start:
                new.y0 = new.y0 + y.start * new.dy
            if y.step:
                new.dy = new.dy * y.step
        return new

    # -------------------------------------------
    # Array2D properties

    @property
    def yindex(self):
        """Positions of the data on the y-axis

        :type: `~astropy.units.Quantity` array
        """
        try:
            return self._yindex
        except AttributeError:
            self._yindex = self.y0 + (
                numpy.arange(self.shape[1]) * self.dy)
            return self._yindex

    @yindex.setter
    def yindex(self, index):
        if isinstance(index, Quantity):
            self._yindex = index
        elif index is None:
            del self.yindex
            return
        else:
            index = Quantity(index, self._default_yunit)
            self._yindex = index
        self.y0 = index[0]
        if index.size:
            self.dy = index[1] - index[0]
        else:
            self.dy = None

    @yindex.deleter
    def yindex(self):
        try:
            del self._yindex
        except AttributeError:
            pass

    @property
    def y0(self):
        """Y-axis value of the first data point

        :type: `~astropy.units.Quantity` scalar
        """
        return self._y0

    @y0.setter
    def y0(self, value):
        if not isinstance(value, Quantity) and value is not None:
            value = Quantity(value, self._default_yunit)
        try:
            y0 = self.y0
        except AttributeError:
            del self.yindex
        else:
            if value is None or self.y0 is None or value != y0:
                del self.yindex
        self._y0 = value

    @y0.deleter
    def y0(self):
        self._y0 = None

    @property
    def dy(self):
        return self._dy

    @dy.setter
    def dy(self, value):
        """Y-axis sample separation

        :type: `~astropy.units.Quantity` scalar
        """
        if not isinstance(value, Quantity) and value is not None:
            value = Quantity(value).to(self.yunit)
        try:
            dy = self.dy
        except AttributeError:
            del self.yindex
        else:
            if value is None or self.dy is None or value != dy:
                del self.yindex
        self._dy = value

    @dy.deleter
    def dy(self):
        self._dy = None

    @property
    def yunit(self):
        """Unit of y-axis index

        :type: `~astropy.units.Unit`
        """
        return self.y0.unit

    @property
    def yspan(self):
        """Y-axis [low, high) segment encompassed by these data

        :type: `~gwpy.segments.Segment`
        """
        from ..segments import Segment
        y0 = self.y0.to(self._default_yunit).value
        dy = self.dy.to(self._default_yunit).value
        return Segment(y0, y0+self.shape[1]*dy)

    # -------------------------------------------
    # numpy.ndarray method modifiers
    # all of these try to return Quantities rather than simple numbers

    def _wrap_function(self, function, *args, **kwargs):
        out = super(Array2D, self)._wrap_function(function, *args, **kwargs)
        if out.ndim == 1:
            # HACK: need to check astropy will always pass axis as first arg
            axis = args[0]
            # return Series
            if axis == 0:
                x0 = self.y0
                dx = self.dy
                xindex = hasattr(self, '_yindex') and self.yindex or None
            else:
                x0 = self.x0
                dx = self.dx
                xindex = hasattr(self, '_xindex') and self.xindex or None
            return Series(out.value, unit=out.unit, x0=x0, dx=dx,
                          channel=out.channel, epoch=self.epoch, xindex=xindex,
                          name='%s %s' % (self.name, function.__name__))
        return out

    def __array_wrap__(self, obj, context=None):
        result = super(Array2D, self).__array_wrap__(obj, context=context)
        try:
            result._xindex = self._xindex
        except AttributeError:
            pass
        try:
            result._yindex = self._yindex
        except AttributeError:
            pass
        return result

    def copy(self, order='C'):
        new = super(Array2D, self).copy(order=order)
        try:
            new._xindex = self._xindex.copy()
        except AttributeError:
            pass
        try:
            new._yindex = self._yindex.copy()
        except AttributeError:
            pass
        return new
    copy.__doc__ = Series.copy.__doc__



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for data input/output in standard GW formats.
"""

from types import FunctionType

from astropy.io.registry import (read, write)

from .. import version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version


def reader(name=None, doc=None):
    """Construct a new unified input/output reader.

    This method is required to create a new copy of the
    :func:`astropy.io.registry.read` with a dynamic docstring.

    Returns
    -------
    read : `function`
        A copy of the :func:`astropy.io.registry.read` function
    """
    func = FunctionType(read.func_code, read.func_globals,
                        name or read.func_name, read.func_defaults,
                        read.func_closure)
    if doc is not None:
        func.__doc__ = doc.strip('\n ')
    return func


def writer():
    """Construct a new unified input/output writeer.

    This method is required to create a new copy of the
    :func:`astropy.io.registry.write` with a dynamic docstring.

    Returns
    -------
    write : `function`
        A copy of the :func:`astropy.io.registry.write` function
    """
    return FunctionType(write.func_code, write.func_globals,
                        write.func_name, write.func_defaults,
                        write.func_closure)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read gravitational-wave data from a cache of files
"""

from astropy.io import registry

from ...io.cache import (identify_cache, identify_cache_file,
                         read_cache_factory)
from .. import (DataQualityFlag, DataQualityDict, SegmentList, SegmentListDict)

# register cache reading
for class_ in [DataQualityFlag, DataQualityDict, SegmentList,
               SegmentListDict]:
    registry.register_reader('lcf', class_, read_cache_factory(class_))
    registry.register_reader('cache', class_, read_cache_factory(class_))
    registry.register_identifier('lcf', class_, identify_cache_file)
    registry.register_identifier('cache', class_, identify_cache)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Calculating the time-dependent coherence between two channels

The standard coherence calculation outputs a frequency series
(`~gwpy.spectrum.Spectrum`) giving a time-averaged measure of coherence.

The `TimeSeries` method :meth:`~TimeSeries.coherence_spectrogram` performs the
same coherence calculation every ``stride``, giving a time-varying coherence
measure.

"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.timeseries'

# First, we import the `TimeSeriesDict`
from gwpy.timeseries import TimeSeriesDict

# and then :meth:`~TimeSeriesDict.get` both data sets:
data = TimeSeriesDict.get(['L1:LSC-SRCL_IN1_DQ', 'L1:LSC-CARM_IN1_DQ'],
                           'Feb 13 2015', 'Feb 13 2015 00:15')

# We can then use the :meth:`~TimeSeries.coherence_spectrogram` method
# of one `TimeSeries` to calcululate the time-varying coherence with
# respect to the other, using a 0.5-second FFT length, with a
# 0.45-second (90%) overlap, with a 8-second stride:
coh = data['L1:LSC-SRCL_IN1_DQ'].coherence_spectrogram(
    data['L1:LSC-CARM_IN1_DQ'], 8, 0.5, 0.45)

# Finally, we can :meth:`~gwpy.spectrogram.Spectrogram.plot` the
# resulting data
plot = coh.plot()
ax = plot.gca()
ax.set_ylabel('Frequency [Hz]')
ax.set_yscale('log')
ax.set_ylim(10, 8000)
ax.set_title('Coherence between SRCL and CARM for L1')
ax.grid(True, 'both', 'both')
plot.add_colorbar(label='Coherence', clim=[0, 1])
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utility module to initialise a kerberos ticket for NDS2 connections

This module provides a lazy-mans python version of the 'kinit'
command-line tool, with internal guesswork using keytabs

See the documentation of the `kinit` function for example usage
"""

import getpass
import os
import sys

try:
    raw_input
except NameError:
    raw_input = input

import re
from subprocess import (PIPE, Popen)

from .. import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

__all__ = ['kinit']


class KerberosError(RuntimeError):
    pass


def which(program):
    """Find full path of executable program

    Parameters
    ----------
    program : `str`
        path of executable name for which to search

    Returns
    -------
    programpath
        the full absolute path of the executable

    Raises
    ------
    ValueError
        if not executable program is found
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    fpath = os.path.split(program)[0]
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    raise ValueError("No executable '%s' found in PATH" % program)


def kinit(username=None, password=None, realm=None, exe=None, keytab=None,
          krb5ccname=None, verbose=False):
    """Initialise a kerberos (krb5) ticket.

    This allows authenticated connections to, amongst others, NDS2
    services.

    Parameters
    ----------
    username : `str`, optional
        name of user, will be prompted for if not given
    password : `str`, optional
        cleartext password of user for given realm, will be prompted for
        if not given
    realm : `str`
        name of realm to authenticate against, defaults to 'LIGO.ORG'
        if not given or parsed from the keytab
    exe : `str`
        path to kinit executable
    keytab : `str`
        path to keytab file. If not given this will be read from the
        ``KRB5_KTNAME`` environment variable. See notes for more details

    Notes
    -----
    If a keytab is given, or is read from the KRB5_KTNAME environment
    variable, this will be used to guess the username and realm, if it
    contains only a single credential

    Examples
    --------
    Example 1: standard user input, with password prompt::

        >>> kinit('albert.einstein')
        Password for albert.einstein@LIGO.ORG:
        Kerberos ticket generated for albert.einstein@LIGO.ORG

    Example 2: extract username and realm from keytab, and use that
    in authentication::

        >>> kinit(keytab='~/.kerberos/ligo.org.keytab')
        Kerberos ticket generated for albert.einstein@LIGO.ORG
    """
    # get kinit path and user keytab
    if exe is None:
        exe = which('kinit')
    if keytab is None:
        keytab = os.environ.get('KRB5_KTNAME', None)
        if keytab is None or not os.path.isfile(keytab):
            keytab = None
    if keytab:
        try:
            principals = parse_keytab(keytab)
        except KerberosError:
            pass
        else:
            # is there's only one entry in the keytab, use that
            if username is None and len(principals) == 1:
                username = principals[0][0]
            # or if the given username is in the keytab, find the realm
            if username in zip(*principals)[0]:
                idx = zip(*principals)[0].index(username)
                realm = principals[idx][1]
            # otherwise this keytab is useless, so remove it
            else:
                keytab = None
    if realm is None:
        realm = 'LIGO.ORG'
    if username is None:
        verbose = True
        username = raw_input("Please provide username for the %s kerberos "
                             "realm: " % realm)
    if not keytab and password is None:
        verbose = True
        password = getpass.getpass(prompt="Password for %s@%s: "
                                          % (username, realm),
                                   stream=sys.stdout)
    if keytab:
        cmd = [exe, '-k', '-t', keytab, '%s@%s' % (username, realm)]
    else:
        cmd = [exe, '%s@%s' % (username, realm)]
    if krb5ccname:
        krbenv = {'KRB5CCNAME': krb5ccname}
    else:
        krbenv = None

    kget = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=krbenv)
    if not keytab:
        kget.communicate(password)
    kget.wait()
    if verbose:
        print("Kerberos ticket generated for %s@%s" % (username, realm))


def parse_keytab(keytab):
    """Read the contents of a KRB5 keytab file, returning a list of
    credentials listed within

    Parameters
    ----------
    keytab : `str`
        path to keytab file
    """
    cmd = ['klist', '-k', '-K', keytab]
    klist = Popen(cmd, stdout=PIPE, stderr=PIPE)
    out = klist.communicate()[0]
    if klist.returncode:
        raise KerberosError("Cannot read keytab '%s'" % keytab)
    principals = []
    for line in out.splitlines():
        try:
            n, principal, _ = re.split('\s+', line.strip(' '), 2)
        except ValueError:
            continue
        else:
            if not n.isdigit():
                continue
            principals.append(principal.split('@'))
    return principals



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read SegmentLists from seg-wizard format ASCII files
"""

from six import string_types

from glue.lal import (CacheEntry, Cache, LIGOTimeGPS)
from glue import segmentsUtils

from astropy.io import registry

from ... import version
from .. import (Segment, SegmentList, DataQualityFlag)
from ...io.cache import file_list

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version


def from_segwizard(f, coalesce=True, gpstype=LIGOTimeGPS, strict=True,
                   nproc=1):
    """Read segments from a segwizard format file into a `SegmentList`
    """
    if nproc != 1:
        return SegmentList.read(f, coalesce=coalesce, gpstype=gpstype,
                                strict=strict, nproc=nproc, format='cache')

    # format list of files and read in serial
    files = file_list(f)
    segs = SegmentList()
    for fp in files:
        if isinstance(fp, file):
            fp = fp.name
        with open(fp, 'r') as fobj:
            segs += SegmentList(map(Segment, segmentsUtils.fromsegwizard(
                fobj, coltype=gpstype, strict=strict)))
    if coalesce:
        segs.coalesce()
    return segs


def flag_from_segwizard(filename, flag=None, coalesce=True, gpstype=float,
                        strict=True, nproc=1):
    if isinstance(flag, DataQualityFlag):
        out = flag
    else:
        out = DataQualityFlag(str(flag))
    out.active = from_segwizard(filename, coalesce=coalesce, gpstype=gpstype,
                                strict=strict, nproc=nproc)
    out.known = out.active
    return out


def identify_segwizard(*args, **kwargs):
    filename = args[3]
    if isinstance(filename, file):
        filename = filename.name
    elif isinstance(filename, CacheEntry):
        filename = filename.path
    if (isinstance(filename, string_types) and
            filename.endswith(('txt', 'dat'))):
        return True
    else:
        return False


def to_segwizard(segs, fobj, header=True, coltype=int):
    """Write the given `SegmentList` to the file object fobj

    Parameters
    ----------
    segs : :class:`~gwpy.segments.segments.SegmentList`
        segmentlist to print
    fobj : `file`, `str`
        open file object, or file path, to write to
    header : `bool`, optional
        print header into the file, default: `True`
    coltype : `type`, optional
        numerical type in which to cast times before printing

    See Also
    --------
    :mod:`glue.segmentsUtils`
        for definition of the segwizard format, and the to/from functions
        used in this GWpy module
    """
    if isinstance(fobj, string_types):
        close = True
        fobj = open(fobj, 'w')
    else:
        close = False
    segmentsUtils.tosegwizard(fobj, segs, header=header, coltype=coltype)
    if close:
        fobj.close()


def flag_to_segwizard(flag, fobj, header=True, coltype=int):
    """Write the given `DataQualityFlag` to the file object fobj

    Parameters
    ----------
    flag : :class:`~gwpy.segments.flag.DataQualityFlag`
        data quality flag to print
    fobj : `file`, `str`
        open file object, or file path, to write to
    header : `bool`, optional
        print header into the file, default: `True`
    coltype : `type`, optional
        numerical type in which to cast times before printing

    Raises
    ------
    ValueError
        if the `~DataQualityFlag.known` segments for `flag` do not mirror
        the `~DataQualityFlag.active` segments. In other words, if recording
        only the `active` segments would discard information.

    Notes
    -----
    In this format, only the
    :attr:`~gwpy.segments.flag.DataQualityFlag.active` segments are
    printed

    See Also
    --------
    :mod:`glue.segmentsUtils`
        for definition of the segwizard format, and the to/from functions
        used in this GWpy module
    """
    if flag.known and flag.known != flag.active:
        raise ValueError("This DataQualityFlag has known segments that do not "
                         "simply match the active ones, meaning the SegWizard "
                         "format cannot preserve these data completely. "
                         "Consider using HDF5 or LIGO_LW XML, otherwise call "
                         "the write() method of the active SegmentList "
                         "directly to write just those segments.")
    to_segwizard(flag.active, fobj, header=header, coltype=coltype)


registry.register_reader('segwizard', DataQualityFlag, flag_from_segwizard)
registry.register_writer('segwizard', DataQualityFlag, flag_to_segwizard)
registry.register_identifier('segwizard', DataQualityFlag, identify_segwizard)

registry.register_reader('segwizard', SegmentList, from_segwizard)
registry.register_writer('segwizard', SegmentList, to_segwizard)
registry.register_identifier('segwizard', SegmentList, identify_segwizard)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Representation of a frequency-series spectrum
"""

import warnings
from copy import deepcopy

from scipy import signal

from astropy import units

from ..data import (Array, Series)
from ..detector import Channel
from ..utils import with_import
from ..utils.docstring import interpolate_docstring


from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org"

__all__ = ['Spectrum']

interpolate_docstring.update({
    'frequency-axis': (
        """f0 : `float`, `~astropy.units.Quantity`, optional, default: `0`
        starting frequency for these data
    df : `float`, `~astropy.units.Quantity`, optional, default: `1`
        frequency resolution for these data
    frequencies : `array-like`
        the complete array of frequencies indexing the data.
        This argument takes precedence over `f0` and `df` so should
        be given in place of these if relevant, not alongside"""),
})


@interpolate_docstring
class Spectrum(Series):
    """A data array holding some metadata to represent a spectrum.

    Parameters
    ----------
    %(Array1)s

    %(frequency-axis)s

    %(Array2)s

    Notes
    -----
    Key methods:

    .. autosummary::

       ~Spectrum.read
       ~Spectrum.write
       ~Spectrum.plot
       ~Spectrum.zpk
    """
    _metadata_slots = Array._metadata_slots + ['f0', 'df']
    _default_xunit = units.Unit('Hz')

    def __new__(cls, data, unit=None, frequencies=None, name=None,
                epoch=None, f0=0, df=1, channel=None,
                **kwargs):
        """Generate a new Spectrum.
        """
        # parse Channel input
        if channel:
            channel = (isinstance(channel, Channel) and channel or
                       Channel(channel))
            name = name or channel.name
            unit = unit or channel.unit
        if frequencies is None and 'xindex' in kwargs:
            frequencies = kwargs.pop('xindex')
        # generate Spectrum
        return super(Spectrum, cls).__new__(cls, data, name=name, unit=unit,
                                            channel=channel, x0=f0, dx=df,
                                            xindex=frequencies, **kwargs)

    # -------------------------------------------
    # Spectrum properties

    f0 = property(Series.x0.__get__, Series.x0.__set__, Series.x0.__delete__,
                  """Starting frequency for this `Spectrum`

                  :type: `~astropy.units.Quantity` scalar
                  """)

    df = property(Series.dx.__get__, Series.dx.__set__, Series.dx.__delete__,
                  """Frequency spacing of this `Spectrum`

                  :type: `~astropy.units.Quantity` scalar
                  """)

    frequencies = property(fget=Series.xindex.__get__,
                           fset=Series.xindex.__set__,
                           fdel=Series.xindex.__delete__,
                           doc="""Series of frequencies for each sample""")

    # -------------------------------------------
    # Spectrum methods

    def plot(self, **kwargs):
        """Display this `Spectrum` in a figure

        All arguments are passed onto the
        :class:`~gwpy.plotter.SpectrumPlot` constructor

        Returns
        -------
        SpectrumPlot
            a new :class:`~gwpy.plotter.SpectrumPlot` rendering
            of this `Spectrum`
        """
        from ..plotter import SpectrumPlot
        return SpectrumPlot(self, **kwargs)

    def zpk(self, zeros, poles, gain):
        """Filter this `Spectrum` by applying a zero-pole-gain filter

        Parameters
        ----------
        zeros : `array-like`
            list of zero frequencies (in Hertz)
        poles : `array-like`
            list of pole frequencies (in Hertz)
        gain : `float`
            DC gain of filter

        Returns
        -------
        spectrum : `Spectrum`
            the frequency-domain filtered version of the input data

        See Also
        --------
        Spectrum.filter
            for details on how a digital ZPK-format filter is applied

        Examples
        --------
        To apply a zpk filter with file poles at 100 Hz, and five zeros at
        1 Hz (giving an overall DC gain of 1e-10)::

            >>> data2 = data.zpk([100]*5, [1]*5, 1e-10)
        """
        return self.filter(zeros, poles, gain)

    def filter(self, *filt, **kwargs):
        """Apply the given filter to this `Spectrum`.

        Recognised filter arguments are converted into the standard
        ``(numerator, denominator)`` representation before being applied
        to this `Spectrum`.

        .. note::

           Unlike the related
           :meth:`TimeSeries.filter <gwpy.timeseries.TimeSeries.filter>`
           method, here all frequency information (e.g. frequencies of
           poles or zeros in a ZPK) is assumed to be in Hertz.

        Parameters
        ----------
        *filt
            one of:

            - :class:`scipy.signal.lti`
            - ``(numerator, denominator)`` polynomials
            - ``(zeros, poles, gain)``
            - ``(A, B, C, D)`` 'state-space' representation

        Returns
        -------
        result : `Spectrum`
            the filtered version of the input `Spectrum`

        See also
        --------
        Spectrum.zpk
            for information on filtering in zero-pole-gain format
        scipy.signal.zpk2tf
            for details on converting ``(zeros, poles, gain)`` into
            transfer function format
        scipy.signal.ss2tf
            for details on converting ``(A, B, C, D)`` to transfer function
            format
        scipy.signal.freqs
            for details on the filtering calculation

        Raises
        ------
        ValueError
            If ``filt`` arguments cannot be interpreted properly
        """
        # parse filter
        if len(filt) == 1 and isinstance(filt[0], signal.lti):
            filt = filt[0]
            a = filt.den
            b = filt.num
        elif len(filt) == 2:
            b, a = filt
        elif len(filt) == 3:
            b, a = signal.zpk2tf(*filt)
        elif len(filt) == 4:
            b, a = signal.ss2tf(*filt)
        else:
            raise ValueError("Cannot interpret filter arguments. Please give "
                             "either a signal.lti object, or a tuple in zpk "
                             "or ba format. See scipy.signal docs for "
                             "details.")
        # parse keyword args
        inplace = kwargs.pop('inplace', False)
        if kwargs:
            raise TypeError("Spectrum.filter() got an unexpected keyword "
                            "argument '%s'" % list(kwargs.keys())[0])
        fresp = abs(signal.freqs(b, a, self.frequencies.value)[1])
        if inplace:
            self.value *= fresp
            return self
        else:
            new = (self.value * fresp).view(type(self))
            new.__dict__ = deepcopy(self.__dict__)
            return new

    def filterba(self, *args, **kwargs):
        warnings.warn("filterba will be removed soon, please use "
                      "Spectrum.filter instead, with the same arguments",
                      DeprecationWarning)
        return self.filter(*args, **kwargs)

    @classmethod
    @with_import('lal')
    def from_lal(cls, lalfs, copy=True):
        """Generate a new `Spectrum` from a LAL `FrequencySeries` of any type
        """
        from ..utils.lal import from_lal_unit
        try:
            unit = from_lal_unit(lalfs.sampleUnits)
        except TypeError:
            unit = None
        channel = Channel(lalfs.name, unit=unit,
                          dtype=lalfs.data.data.dtype)
        return cls(lalfs.data.data, channel=channel, f0=lalfs.f0,
                   df=lalfs.deltaF, epoch=float(lalfs.epoch),
                   dtype=lalfs.data.data.dtype, copy=copy)

    @with_import('lal')
    def to_lal(self):
        """Convert this `Spectrum` into a LAL FrequencySeries.

        Returns
        -------
        lalspec : `FrequencySeries`
            an XLAL-format FrequencySeries of a given type, e.g.
            :lal:`REAL8FrequencySeries`

        Notes
        -----
        Currently, this function is unable to handle unit string
        conversion.
        """
        from ..utils.lal import (LAL_TYPE_STR_FROM_NUMPY, to_lal_unit)
        typestr = LAL_TYPE_STR_FROM_NUMPY[self.dtype.type]
        try:
            unit = to_lal_unit(self.unit)
        except TypeError:
            unit = lal.lalDimensionlessUnit
        create = getattr(lal, 'Create%sFrequencySeries' % typestr.upper())
        lalfs = create(self.name, lal.LIGOTimeGPS(self.epoch.gps),
                       self.f0.value, self.df.value, unit, self.size)
        lalfs.data.data = self.value
        return lalfs

    @classmethod
    def from_pycbc(cls, fs):
        """Convert a `pycbc.types.frequencyseries.FrequencySeries` into
        a `Spectrum`

        Parameters
        ----------
        fs : `pycbc.types.frequencyseries.FrequencySeries`
            the input PyCBC `~pycbc.types.frequencyseries.FrequencySeries`
            array

        Returns
        -------
        spectrum : `Spectrum`
            a GWpy version of the input frequency series
        """
        return cls(fs.data, f0=0, df=fs.delta_f, epoch=fs.epoch)

    @with_import('pycbc.types')
    def to_pycbc(self, copy=True):
        """Convert this `Spectrum` into a
        `pycbc.types.frequencyseries.FrequencySeries`

        Parameters
        ----------
        copy : `bool`, optional, default: `True`
            if `True`, copy these data to a new array

        Returns
        -------
        frequencyseries : `pycbc.types.frequencyseries.FrequencySeries`
            a PyCBC representation of this `Spectrum`
        """
        return types.FrequencySeries(self.data, delta_f=self.df.to('Hz').value,
                                     epoch=self.epoch.gps, copy=copy)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting an event histogram

I would like to study the snr distribution of event triggers generated by
the ExcessPower gravitational-wave burst detection algorithm,
over a small stretch of data.

The data from which these events were generated are a simulation of Gaussian
noise with the Advanced LIGO design spectrum, and so don't actually contain
any real gravitational waves, but will help tune the algorithm to improve
detection of future, real signals.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.table.lsctables'

# First, we import the `SnglBurstTable`, an XML-based data holder for single-detector ('Sngl') gravitational-wave burst-like events:
from gwpy.table.lsctables import SnglBurstTable

# We can read a set of events using the :meth:`~SnglBurstTable.read` method:
events = SnglBurstTable.read('../../gwpy/tests/data/H1-LDAS_STRAIN-968654552-10.xml.gz')

# and can generate a new `~gwpy.plotter.HistogramPlot` using the :meth:`~SnglBurstTable.hist` instance method
plot = events.hist('snr', weights=1/10., log=True, logbins=True, histtype='stepfilled', cumulative=-1)
plot.set_xlabel('Signal-to-noise ratio (SNR)')
plot.set_ylabel('Rate [Hz]')
plot.set_title('LHO event triggers for GW100916')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read data from gravitational-wave frames using frameCPP.
"""

from __future__ import division
import __builtin__

import numpy

from .... import version
from ....io.cache import (CacheEntry, file_list)
from ....segments import Segment
from ....utils import (gprint, with_import)
from ... import (TimeSeries, TimeSeriesDict)

from . import channel_dict_kwarg

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

# get frameCPP path
try:
    from LDAStools.frameCPP import FrVect
except ImportError:
    DEPENDS = 'frameCPP'
    try:
        from frameCPP import FrVect
    except ImportError:
        FrVect = None
else:
    DEPENDS = 'LDAStools.frameCPP'

# get frameCPP type mapping
if FrVect is not None:
    NUMPY_TYPE_FROM_FRVECT = {
        FrVect.FR_VECT_C: numpy.int8,
        FrVect.FR_VECT_2S: numpy.int16,
        FrVect.FR_VECT_8R: numpy.float64,
        FrVect.FR_VECT_4R: numpy.float32,
        FrVect.FR_VECT_4S: numpy.int32,
        FrVect.FR_VECT_8S: numpy.int64,
        FrVect.FR_VECT_8C: numpy.complex64,
        FrVect.FR_VECT_16C: numpy.complex128,
        FrVect.FR_VECT_STRING: numpy.string_,
        FrVect.FR_VECT_2U: numpy.uint16,
        FrVect.FR_VECT_4U: numpy.uint32,
        FrVect.FR_VECT_8U: numpy.uint64,
    }
# default to LDASTools v2.2.1
else:
    NUMPY_TYPE_FROM_FRVECT = {
        0: numpy.int8,
        1: numpy.int16,
        2: numpy.float64,
        3: numpy.float32,
        4: numpy.int32,
        5: numpy.int64,
        6: numpy.complex64,
        7: numpy.complex128,
        8: numpy.string_,
        9: numpy.uint16,
        10: numpy.uint32,
        11: numpy.uint64,
        12: numpy.uint8,
    }


@with_import(DEPENDS)
def read_timeseriesdict(source, channels, start=None, end=None, type=None,
                        dtype=None, resample=None, verbose=False,
                        _SeriesClass=TimeSeries):
    """Read the data for a list of channels from a GWF data source.

    Parameters
    ----------
    source : `str`, :class:`glue.lal.Cache`, `list`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache`, `list` : contiguous list of frame paths

    channels : `list`
        list of channel names (or `Channel` objects) to read from frame.
    start : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        start GPS time of desired data.
    end : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        end GPS time of desired data.
    type : `str`
        type of channel, one of ``'adc'`` or ``'proc'``.
    dtype : `numpy.dtype`, `str`, `type`, or `dict`
        numeric data type for returned data, e.g. `numpy.float`, or
        `dict` of (`channel`, `dtype`) pairs
    resample : `float`, optional
        rate of samples per second at which to resample input TimeSeries.
    verbose : `bool`, optional
        print verbose output.

    Returns
    -------
    dict : :class:`~gwpy.timeseries.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs

    Notes
    -----
    If reading from a list, or cache, or framefiles, the frames contained
    must be contiguous and sorted in chronological order for this function
    to return without exception.

    Raises
    ------
    ValueError
        if reading from an unsorted, or discontiguous cache of files
    """
    frametype = None
    # parse input source
    filelist = file_list(source)
    try:
        frametype = CacheEntry.from_T050017(filelist[0]).description
    except ValueError:
        frametype = None
    # parse resampling
    resample = channel_dict_kwarg(resample, channels, (int,))
    if resample is None:
        raise ValueError("Cannot parse `resample` request, please review "
                         "documentation for that argument")
    # parse type
    type = channel_dict_kwarg(type, channels, (str,))
    if resample is None:
        raise ValueError("Cannot parse channel `type` request, please review "
                         "documentation for that argument")
    # parse dtype
    dtype = channel_dict_kwarg(dtype, channels, (str, __builtin__.type),
                               astype=numpy.dtype)
    if dtype is None:
        raise ValueError("Cannot parse `dtype` request, please review "
                         "documentation for that argument")
    # read each individually and append
    N = len(filelist)
    if verbose:
        if not isinstance(verbose, (unicode, str)):
            verbose = ''
        gprint("%sReading %d channels from frames... 0/%d (0.00%%)\r"
               % (verbose, len(channels), N), end='')
    out = TimeSeriesDict()
    for i, fp in enumerate(filelist):
        # read frame
        new = _read_frame(fp, channels, start=start, end=end, ctype=type,
                          dtype=dtype, _SeriesClass=_SeriesClass)
        ## get channel type for next frame (means we only query the TOC once)
        if not i:
            for channel, ts in new.iteritems():
                type[channel] = ts.channel._ctype
        # store
        out.append(new, copy=False)
        if verbose is not False:
            gprint("%sReading %d channels from frames... %d/%d (%.1f%%)\r"
                   % (verbose, len(channels), i+1, N, (i+1)/N * 100), end='')
    if verbose is not False:
        gprint("%sReading %d channels from frames... %d/%d (100.0%%)"
               % (verbose, len(channels), N, N))
    # finalise
    for channel, ts in out.iteritems():
        ts.channel.frametype = frametype
        # resample data
        if resample is not None and channel in resample:
            out[channel] = out[channel].resample(resample[channel])
        # crop data
        if start is not None or end is not None:
            out[channel] = out[channel].crop(start=start, end=end)
        # copy into fresh memory if needed
        out[channel] = numpy.require(out[channel], requirements=['O'])
    return out


def _read_frame(framefile, channels, start=None, end=None, ctype=None,
                dtype=None, _SeriesClass=TimeSeries):
    """Internal function to read data from a single frame.

    All users should be using the wrapper `read_timeseriesdict`.

    Parameters
    ----------
    framefile : `str`, :class:`~glue.lal.CacheEntry`
        path to GWF-format frame file on disk.
    channels : `list`
        list of channels to read.
    start : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        start GPS time of desired data.
    end : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        end GPS time of desired data.
    ctype : `str`, optional
        channel data type to read, one of: ``'adc'``, ``'proc'``.
    dtype : `numpy.dtype`, `str`, `type`, `dict`
    _SeriesClass : `type`, optional
        class object to use as the data holder for a single channel,
        default is :class:`~gwpy.timeseries.TimeSeries`

    Returns
    -------
    dict : :class:`~gwpy.timeseries.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs
    """
    if isinstance(channels, (unicode, str)):
        channels = channels.split(',')

    # construct span segment
    span = Segment(start is not None and start or -numpy.inf,
                   end is not None and end or numpy.inf)

    # open file
    if isinstance(framefile, CacheEntry):
        fp = framefile.path
    else:
        fp = framefile
    stream = frameCPP.IFrameFStream(fp)

    # interpolate frame epochs from CacheEntry
    try:
        nframe = int(stream.GetNumberOfFrames())
    except (AttributeError, ValueError):
        nframe = None
    if isinstance(framefile, CacheEntry) and nframe == 1:
        epochs = [float(framefile.segment[0])]
    else:
        epochs = None

    # load table of contents if needed
    if epochs is None or not ctype:
        toc = stream.GetTOC()
    # get list of frame epochs
    if epochs is None:
        epochs = toc.GTimeS
    # work out channel types
    if not ctype:
        try:
            adcs = toc.GetADC().keys()
        except AttributeError:
            adcs = toc.GetADC()
        try:
            procs = toc.GetProc().keys()
        except AttributeError:
            procs = toc.GetProc()
        ctype = {}
        for channel in channels:
            name = str(channel)
            if name in adcs:
                ctype[channel] = 'adc'
            elif name in procs:
                ctype[channel] = 'proc'
            else:
                raise ValueError("Channel %s not found in frame table of "
                                 "contents" % name)

    # set output
    out = TimeSeriesDict()
    for channel in channels:
        name = str(channel)
        read_ = getattr(stream, 'ReadFr%sData' % ctype[channel].title())
        ts = None
        i = 0
        dtype_ = dtype.get(channel, None)
        while True:
            try:
                data = read_(i, name)
            except IndexError:
                break
            offset = data.GetTimeOffset()
            thisepoch = epochs[i] + offset
            try:
                thisspan = Segment(thisepoch, thisepoch + data.GetTRange())
            except AttributeError:
                pass
            else:
                if not thisspan.intersects(span):
                    i += 1
                    continue
            for vect in data.data:
                arr = vect.GetDataArray()
                if isinstance(arr, buffer):
                   arr = numpy.frombuffer(
                       arr, dtype=NUMPY_TYPE_FROM_FRVECT[vect.GetType()])
                dx = vect.GetDim(0).dx
                if ts is None:
                    # create array
                    unit = vect.GetUnitY() or None
                    ts = numpy.require(
                        _SeriesClass(arr, epoch=thisepoch, dx=dx, name=name,
                                     channel=channel, unit=unit, dtype=dtype_,
                                     copy=False), requirements=['O'])
                    if not ts.channel.dtype:
                        ts.channel.dtype = arr.dtype
                    ts.channel._ctype = ctype[channel]
                elif arr.dtype != ts.dtype:
                    ts.append(arr.astype(dtype_))
                else:
                    ts.append(arr)
            i += 1
        if ts is None:
            raise ValueError("Channel '%s' not found in frame '%s'"
                             % (str(channel), fp))
        else:
            out[channel] = ts

    return out
